# lsqfitgp/tests/test_linalg.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import re
import sys
import abc
import inspect
import functools

import jax
from jax import numpy as jnp
from jax import test_util
from jax.scipy import linalg as jlinalg
import numpy as np
from scipy import linalg, stats
import gvar
from pytest import mark
import pytest

import util

sys.path.insert(0, '.')
from lsqfitgp import _linalg, _kernels

TRYAGAIN = True

# TODO rewrite many comparisons to check for closeness of inputs with inverse
# operation applied to solution in 2-norm instead of comparing solutions
# computed in different ways

# TODO since the decompositions are pseudoinverses, I should do all tests on
# singular matrices too instead of just for some of them in test_degenerate. To
# make a degenerate matrix, multiply the current matrix factories by a random
# singular projector. To do:
#   - fix and document pinv behavior in Decomposition
#   - decide which properties to check (just the first MP identity? All four?)

# TODO remove some redundant tests to speed up the suite (_matrix and _vec
# variants?)

jax.config.update('jax_default_matmul_precision', jax.lax.Precision.HIGHEST)

s1, s2 = np.random.SeedSequence(202302061416).spawn(2)
rng = np.random.default_rng(s1)
np.random.seed(s2.generate_state(1))

def randortho(n):
    if n > 1:
        return stats.ortho_group.rvs(n, random_state=rng)
    else:
        # stats.ortho_group does not support n < 2
        return np.atleast_2d(2 * rng.integers(2) - 1)

class DecompTestABC(abc.ABC):

    @property
    @abc.abstractmethod
    def decompclass(self):
        """ Decomposition subclass to be tested """
        pass
    
    @property
    @abc.abstractmethod
    def lowrank(self):
        """ Boolean indicating if the tests matrices are singular """
        pass
        
    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)
        
        cls.matjac = jax.jacfwd(cls.mat, 1)
        cls.mathess = jax.jacfwd(cls.matjac, 1)
            
        # wrap all methods to try again in case of failure, to hedge against bad
        # random matrices. vars() only gives methods overwritten by a subclass,
        # so we do not get all members, such that subclasses inherit xfails,
        # which would be hidden by the decorator
        if TRYAGAIN:
            for name, meth in vars(cls).items():
                if name.startswith('test_'):
                        setattr(cls, name, util.tryagain(meth, method=True))
        
        def setattr_ifmis(cls, name, value):
            prev = getattr(cls, name, None)
            sup = getattr(DecompTestABC, name, prev)
            if not hasattr(cls, name) or sup is prev:
                setattr(cls, name, value)
        
        # configure automatically the subclass if it's still abstract
        name = cls.__name__
        if name.startswith('Test') and inspect.isabstract(cls):
            name = name[4:]
            if name.endswith('_LowRank'):
                name = name[:-8]
                cls.lowrank = True
            else:
                cls.lowrank = False
            if '_' in name:
                comp, name = name.split('_')
                setattr_ifmis(cls, 'compositeclass', getattr(_linalg, comp))
                setattr_ifmis(cls, 'subdecompclass', getattr(_linalg, name))
            else:
                setattr_ifmis(cls, 'decompclass', getattr(_linalg, name))

class DecompTestBase(DecompTestABC):
    """
    
    Restrictions:
    
    - self.decompclass can be applied only on a matrix generated either with
      self.mat or self.randsymmat
    
    - in each test method, use only one of self.mat or self.randsymmat
    
    """
    
    sizes = [10, 3, 2, 1]
    
    def randsize(self):
        return rng.integers(1, 11)
    
    _rank = functools.cached_property(lambda self: {})
    def rank(self, n):
        # cache the rank because it needs to stay consistent between invocations
        # of parametric test matrix constructors
        return self._rank.setdefault(n, 1 + rng.integers(n) if self.lowrank else n)
    
    # TODO unificate randsymmat and mat by generating a cached random orthogonal
    # transformation and using a "pseudocasual" nonlinear function of the
    # parameter for the eigenvalues. To generate a random matrix, pass a random
    # parameter. This simplifies the composite decompositions. randsymmat can
    # become a wrapper of mat. Even better for the composite: generate directly
    # the decomposition, i.e., replace decompclass with decomp(n, s) and
    # randdecomp(n), since now I can get the matrix from the decomposition with
    # .matrix(). => Problem: if I fix the orthogonal transf, ∂K will have the
    # same kernel as K, which is not general. I could add a rotation around a
    # random axis after the orthogonal transformation.
        
    def randsymmat(self, n, *, rank=None):
        """ return random nxn symmetric matrix, well conditioned in the
        orthogonal complement to the null space """
        if rank is None:
            rank = self.rank(n)
        eigvals = rng.uniform(1e-2, 1e2, size=rank)
        O = randortho(n)
        O = O[:, :rank]
        K = (O * eigvals) @ O.T
        util.assert_close_matrices(K, K.T, rtol=1e-15)
        return K
    
    def mat(self, s, n, *, rank=None):
        if rank is None:
            rank = self.rank(n)
        x = np.arange(n)
        x[:n - rank + 1] = x[:1]
        return np.pi * jnp.exp(-1/2 * (x[:, None] - x[None, :]) ** 2 / s ** 2)
        # the π factor is for tests that pass only if the diagonal is 1
        # TODO add some nontrivial dependence on s when n = 1
    
    def randvec(self, n):
        return rng.standard_normal(n)
    
    def randmat(self, m, n=None):
        if n is None:
            n = self.randsize()
        return rng.standard_normal((m, n))
    
    def quad_or_solve(self, K, b=None, c=None, return_rank=False):
        U, s, Vh = linalg.svd(K)
        eps = len(K) * s[0] * np.finfo(K.dtype).eps
        cond = s > eps
        sinv = np.diag(np.where(cond, 1 / np.where(s, s, 1), 0))
        lpd = np.sum(np.log(s[cond]))
        ld = lpd + np.count_nonzero(~cond) * np.log(eps)
        rank = np.count_nonzero(cond)

        if b is None:
            sol = lpd
        elif c is None:
            sol = Vh.T @ (sinv @ (U.T @ b))
        else:
            sol = (Vh @ b).T @ sinv @ (U.T @ c)

        if return_rank:
            return sol, rank
        else:
            return sol
    
    def logdet(self, K):
        return self.quad_or_solve(K)

    def solve(self, K, b):
        """ K^+ b """
        return self.quad_or_solve(K, b)
    
    def quad(self, K, b, c=None):
        """ b^T K^+ b or b^T K^+ c """
        return self.quad_or_solve(K, b, b if c is None else c)
    
    @classmethod
    def clskey(cls, mapping, default):
        for k, v in mapping.items():
            if re.search(k, cls.__name__, flags=re.IGNORECASE):
                return v
        return default
    
    def check_solve(self, bgen, jit=False):
        fun = lambda K, b: self.decompclass(K).solve(b)
        for n in self.sizes:
            K = self.randsymmat(n)
            b = bgen(len(K))
            result = fun(K, b)
            if jit:
                result2 = jax.jit(fun)(K, b)
                util.assert_close_matrices(result2, result, rtol=self.clskey({
                    r'pinv_chol': 1e-8,
                    r'pinv2_chol': 1e-9,
                }, 1e-11))
            else:
                sol = self.solve(K, b)
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'pinv_chol': 1e-3,
                    r'pinv2_chol': 1e-7,
                    r'TestSandwichSVD_EigCutFullRank': 1e-10,
                }, 1e-11), atol=1e-15)
    
    def check_solve_jac(self, bgen, jacfun, jit=False, hess=False, da=False, rtol=1e-7):
        # TODO sometimes the jacobian of fun is practically zero. Why? This
        # gives problems because it needs an higher absolute tolerance. =>
        # because in 1x1 matrices there is only the diagonal, which does not
        # depend on s (but this can't be that problem, because it's exactly 0)
        def fun(s, n, b):
            K = self.mat(s, n)
            return self.decompclass(K, direct_autodiff=da).solve(b)
        funjac = jacfun(fun)
        for n in self.sizes:
            s = np.exp(rng.uniform(-1, 1))
            b = bgen(n)
            result = funjac(s, n, b)
            if jit:
                funjacjit = jax.jit(funjac, static_argnums=1)
                result2 = funjacjit(s, n, b)
                util.assert_close_matrices(result2, result, rtol=self.clskey({
                    r'pinv2_chol': 1e-4,
                    r'pinv_chol': 1e-4,
                }, rtol))
                continue
            K = self.mat(s, n)
            dK = self.matjac(s, n)
            KdK = self.solve(K, dK)
            Kb = self.solve(K, b)
            if not hess:
                # TODO use correct formulas for pseudoinverses
                # -K^-1 dK K^-1 b
                sol = -KdK @ Kb
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'woodbury[^2]': 1e-3,
                    r'woodbury2': 1e-3,
                    r'pinv(2|)_chol': 1e-4,
                    r'chol': 1e-6,
                    r'sandwichsvd': 1e-7,
                }, 1e-8))
            else:
                # TODO use correct formulas for pseudoinverses
                #  K^-1 dK K^-1 dK K^-1 b   +
                # -K^-1 d2K K^-1 b          +
                #  K^-1 dK K^-1 dK K^-1 b
                d2K = self.mathess(s, n)
                sol = 2 * KdK @ KdK @ Kb - self.solve(K, d2K) @ Kb
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'pinv_chol': 1e-4,
                    r'woodbury2': 1e-4,
                    r'woodbury[^2]': 1e-5,
                    r'chol': 1e-6,
                }, 1e-7))
    
    def test_solve_vec(self):
        self.check_solve(self.randvec)
    
    def test_solve_vec_jac_rev(self):
        self.check_solve_jac(self.randvec, jax.jacrev)

    def test_solve_vec_jac_fwd(self):
        self.check_solve_jac(self.randvec, jax.jacfwd)

    def test_solve_matrix(self):
        self.check_solve(self.randmat)
    
    def test_solve_matrix_jac_rev(self):
        self.check_solve_jac(self.randmat, jax.jacrev)
    
    def test_solve_matrix_jac_fwd(self):
        self.check_solve_jac(self.randmat, jax.jacfwd)
    
    def test_solve_matrix_jac_fwd_da(self):
        self.check_solve_jac(self.randmat, jax.jacfwd, da=True)

    def test_solve_vec_jit(self):
        self.check_solve(self.randvec, True)
    
    def test_solve_vec_jac_rev_jit(self):
        self.check_solve_jac(self.randvec, jax.jacrev, True)

    def test_solve_vec_jac_fwd_jit(self):
        self.check_solve_jac(self.randvec, jax.jacfwd, True, rtol=self.clskey({
            r'pinv2_chol': 1e-3,
        }, 1e-6))

    def test_solve_matrix_jit(self):
        self.check_solve(self.randmat, True)
    
    def test_solve_matrix_jac_rev_jit(self):
        self.check_solve_jac(self.randmat, jax.jacrev, True)
    
    def test_solve_matrix_jac_fwd_jit(self):
        self.check_solve_jac(self.randmat, jax.jacfwd, True)
    
    def test_solve_matrix_hess_fwd_fwd(self):
        self.check_solve_jac(self.randmat, lambda f: jax.jacfwd(jax.jacfwd(f)), hess=True)

    def test_solve_matrix_hess_fwd_rev(self):
        self.check_solve_jac(self.randmat, lambda f: jax.jacfwd(jax.jacrev(f)), hess=True)

    def test_solve_matrix_hess_da(self):
        self.check_solve_jac(self.randmat, lambda f: jax.jacfwd(jax.jacrev(f)), hess=True, da=True)

    def solve_matrix_jac_matrix(self, jacfun):
        def fun(s, n, b, A):
            K = self.mat(s, n)
            return A @ self.decompclass(K).solve(b)
        funjac = jacfun(fun)
        for n in self.sizes:
            s = np.exp(rng.uniform(-1, 1))
            K = self.mat(s, n)
            dK = self.matjac(s, n)
            b = self.randmat(n)
            A = self.randmat(n).T
                # TODO use correct formulas for pseudoinverses
            sol = -A @ self.solve(K.T, self.solve(K, dK).T).T @ b
            result = funjac(s, n, b, A)
            util.assert_close_matrices(sol, result, rtol=self.clskey({
                r'pinv_chol': 1e-2,
                r'pinv2_chol': 1e-3,
            }, 1e-4))
        
    def test_solve_matrix_jac_rev_matrix(self):
        self.solve_matrix_jac_matrix(jax.jacrev)

    def test_solve_matrix_jac_fwd_matrix(self):
        self.solve_matrix_jac_matrix(jax.jacfwd)

    def randvecgvar(self, n):
        mean = self.randvec(n)
        xcov = np.linspace(0, 3, n)
        cov = np.exp(-(xcov.reshape(-1, 1) - xcov.reshape(1, -1)) ** 2)
        return gvar.gvar(mean, cov)
    
    def test_quad_vec_vec_gvar(self):
        for n in self.sizes:
            K = self.randsymmat(n)
            b = self.randvec(n)
            c = self.randvecgvar(n)
            invK = self.solve(K, np.eye(len(K)))
            sol = b.T @ invK @ c
            result = self.decompclass(K).quad(b, c)
            util.assert_similar_gvars(sol, result, rtol=self.clskey({
                r'pinv_chol': 1e-3,
                r'pinv2_chol': 1e-7,
            }, 1e-10), atol=1e-15)

    def test_quad_matrix_vec_gvar(self):
        for n in self.sizes:
            K = self.randsymmat(n)
            b = self.randvec(n * 3).reshape(n, 3)
            c = self.randvecgvar(n)
            invK = self.solve(K, np.eye(len(K)))
            sol = b.T @ invK @ c
            result = self.decompclass(K).quad(b, c)
            util.assert_similar_gvars(sol, result, rtol=self.clskey({
                r'pinv2_chol': 1e-6,
                r'pinv_chol': 1e-8,
                r'sandwichqr': 1e-10,
            }, 1e-11), atol=1e-15)
    
    def check_quad(self, bgen, cgen=lambda n: None, jit=False):
        fun = lambda K, b, c: self.decompclass(K).quad(b, c)
        for n in self.sizes:
            K = self.randsymmat(n)
            b = bgen(len(K))
            c = cgen(len(K))
            result = fun(K, b, c)
            if jit:
                funjit = jax.jit(fun)
                result2 = funjit(K, b, c)
                util.assert_close_matrices(result2, result, rtol=self.clskey({
                    r'pinv2_chol': 1e-7,
                    r'pinv_chol': 1e-10,
                }, 1e-11))
            else:
                sol = self.quad(K, b, c)
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'pinv_chol': 1e-6,
                }, 1e-8))
    
    def check_quad_jac(self, jacfun, bgen, cgen=lambda n: None, jit=False, hess=False, da=False, stopg=False):
        def fun(s, n, b, c):
            K = self.mat(s, n)
            return self.decompclass(K, direct_autodiff=da, stop_hessian=stopg).quad(b, c)
        fungrad = jacfun(fun)
        fungradjit = jax.jit(fungrad, static_argnums=1)
        for n in self.sizes:
            s = np.exp(rng.uniform(-1, 1))
            b = bgen(n)
            c = cgen(n)
            result = fungrad(s, n, b, c)
            if jit:
                result2 = fungradjit(s, n, b, c)
                util.assert_close_matrices(result2, result, rtol=self.clskey({
                    r'pinv_chol': 1e-2,
                    r'pinv2_chol': 1e-3,
                    r'woodbury[^2]': 1e-4,
                }, 1e-7))
                continue
            if c is None:
                c = b
            K = self.mat(s, n)
            dK = self.matjac(s, n)
            KdK = self.solve(K, dK)
            Kc = self.solve(K, c)
            if not hess:
                # TODO use correct formulas for pseudoinverses
                # b.T K^-1 c
                # -b.T K^-1 dK K^-1 c
                sol = -b.T @ KdK @ Kc
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'pinv_chol': 1e-2,
                    r'woodbury[^2]': 1e-2,
                    r'pinv2_chol': 1e-3,
                    r'chol': 1e-6,
                    r'woodbury2': 1e-7,
                }, 1e-8))
            else:
                # TODO use correct formulas for pseudoinverses
                #  b.T K^-1 dK K^-1 dK K^-1 c   +
                # -b.T K^-1 d2K K^-1 c          +
                #  b.T K^-1 dK K^-1 dK K^-1 c
                sol = 2 * b.T @ KdK @ KdK @ Kc
                if not stopg:
                    d2K = self.mathess(s, n)
                    sol -= b.T @ self.solve(K, d2K) @ Kc
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'pinv_chol': 1e-2,
                    r'pinv2_chol': 1e-4,
                    r'woodbury[^2]': 1e-3,
                }, 1e-6))
    
    def test_quad_vec(self):
        self.check_quad(self.randvec)
    
    def test_quad_vec_vec(self):
        self.check_quad(self.randvec, self.randvec)
    
    def test_quad_matrix(self):
        self.check_quad(self.randmat)
    
    def test_quad_matrix_matrix(self):
        self.check_quad(self.randmat, self.randmat)
    
    def test_quad_vec_jit(self):
        self.check_quad(self.randvec, jit=True)
    
    def test_quad_vec_vec_jit(self):
        self.check_quad(self.randvec, self.randvec, True)
    
    def test_quad_matrix_jit(self):
        self.check_quad(self.randmat, jit=True)
    
    def test_quad_matrix_matrix_jit(self):
        self.check_quad(self.randmat, self.randmat, True)
    
    def test_quad_vec_jac_rev(self):
        self.check_quad_jac(jax.jacrev, self.randvec)
    
    def test_quad_vec_jac_fwd(self):
        self.check_quad_jac(jax.jacfwd, self.randvec)
    
    def test_quad_matrix_jac_rev(self):
        self.check_quad_jac(jax.jacrev, self.randmat)
    
    def test_quad_matrix_jac_fwd(self):
        self.check_quad_jac(jax.jacfwd, self.randmat)
    
    def test_quad_matrix_matrix_jac_fwd(self):
        self.check_quad_jac(jax.jacfwd, self.randmat, self.randmat)
    
    def test_quad_matrix_matrix_jac_fwd_da(self):
        self.check_quad_jac(jax.jacfwd, self.randmat, self.randmat, da=True)
    
    def test_quad_vec_jac_rev_jit(self):
        self.check_quad_jac(jax.jacrev, self.randvec, jit=True)
    
    def test_quad_vec_jac_fwd_jit(self):
        self.check_quad_jac(jax.jacfwd, self.randvec, jit=True)
    
    def test_quad_matrix_jac_rev_jit(self):
        self.check_quad_jac(jax.jacrev, self.randmat, jit=True)
    
    def test_quad_matrix_jac_fwd_jit(self):
        self.check_quad_jac(jax.jacfwd, self.randmat, jit=True)
    
    def test_quad_matrix_matrix_jac_fwd_jit(self):
        self.check_quad_jac(jax.jacfwd, self.randmat, self.randmat, jit=True)

    def test_quad_matrix_matrix_hess_fwd_rev(self):
        self.check_quad_jac(lambda f: jax.jacfwd(jax.jacrev(f)), self.randmat, self.randmat, hess=True)
    
    def test_quad_matrix_matrix_hess_fwd_fwd(self):
        self.check_quad_jac(lambda f: jax.jacfwd(jax.jacfwd(f)), self.randmat, self.randmat, hess=True)
    
    def test_quad_matrix_matrix_hess_fwd_fwd_stopg(self):
        self.check_quad_jac(lambda f: jax.jacfwd(jax.jacfwd(f)), self.randmat, self.randmat, hess=True, stopg=True)
    
    def test_quad_matrix_matrix_hess_da(self):
        self.check_quad_jac(lambda f: jax.jacfwd(jax.jacrev(f)), self.randmat, self.randmat, hess=True, da=True)
    
    def check_logdet(self, jit=False):
        fun = lambda K: self.decompclass(K).logdet()
        funjit = jax.jit(fun)
        for n in self.sizes:
            K = self.randsymmat(n)
            result = fun(K)
            if jit:
                result2 = funjit(K)
                util.assert_close_matrices(result2, result, rtol=self.clskey({
                    r'pinv_chol': 1e-2,
                    r'pinv2_chol': 1e-7,
                }, 1e-14))
            else:
                sol = self.logdet(K)
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'pinv2_chol': 1e-2,
                    r'pinv_chol': 1e-7,
                }, 1e-11))
    
    def check_logdet_jac(self, jacfun, jit=False, hess=False, num=False, da=False, stopg=False):
        def fun(s, n):
            K = self.mat(s, n)
            return self.decompclass(K, direct_autodiff=da, stop_hessian=stopg).logdet()
        fungrad = jacfun(fun)
        fungradjit = jax.jit(fungrad, static_argnums=1)
        for n in self.sizes:
            s = np.exp(rng.uniform(-1, 1))
            if num:
                order = 2 if hess else 1
                test_util.check_grads(lambda s: fun(s, n), (s,), order=order)
                continue
            result = fungrad(s, n)
            if jit:
                result2 = fungradjit(s, n)
                util.assert_close_matrices(result2, result, rtol=self.clskey({
                    r'pinv_chol': 1e-3,
                    r'woodbury2': 1e-6,
                    r'pinv2_chol': 1e-7,
                }, 1e-9), atol=1e-30)
                continue
            K = self.mat(s, n)
            dK = self.matjac(s, n)
            KdK = self.solve(K, dK)
            if not hess:
                # tr(K^-1 dK)
                sol = np.trace(KdK)
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'pinv2_chol': 1e-3,
                    r'pinv_chol': 1e-4,
                    r'TestWoodbury2_EigCutFullRank': 1e-4,
                }, 1e-7), atol=1e-20)
            else:
                # tr(-K^-1 dK K^-1 dK + K d2K)
                sol = -np.trace(KdK @ KdK)
                if not stopg:
                    d2K = self.mathess(s, n)
                    Kd2K = self.solve(K, d2K)
                    sol += np.trace(Kd2K)
                util.assert_close_matrices(result, sol, rtol=self.clskey({
                    r'pinv_chol': 1e-2,
                    r'pinv2_chol': 1e-3,
                    r'woodbury[^2]': 1e-2,
                }, 1e-5), atol=1e-60)
    
    def test_logdet(self):
        self.check_logdet()
    
    def test_logdet_jac_rev(self):
        self.check_logdet_jac(jax.jacrev)
    
    def test_logdet_jac_fwd(self):
        self.check_logdet_jac(jax.jacfwd)
        
    def test_logdet_hess_fwd_fwd(self):
        self.check_logdet_jac(lambda f: jax.jacfwd(jax.jacfwd(f)), hess=True)

    def test_logdet_hess_fwd_fwd_stopg(self):
        self.check_logdet_jac(lambda f: jax.jacfwd(jax.jacfwd(f)), hess=True, stopg=True)

    def test_logdet_hess_fwd_rev(self):
        self.check_logdet_jac(lambda f: jax.jacfwd(jax.jacrev(f)), hess=True)

    def test_logdet_hess_da(self):
        self.check_logdet_jac(lambda f: jax.jacfwd(jax.jacrev(f)), hess=True, da=True)

    def test_logdet_jit(self):
        self.check_logdet(True)
    
    def test_logdet_jac_rev_jit(self):
        self.check_logdet_jac(jax.jacrev, True)
    
    def test_logdet_jac_fwd_jit(self):
        self.check_logdet_jac(jax.jacfwd, True)

    def test_inv(self):
        for n in self.sizes:
            K = self.randsymmat(n)
            sol = self.solve(K, np.eye(n))
            result = self.decompclass(K).inv()
            util.assert_close_matrices(result, sol, rtol=self.clskey({
                r'pinv2_chol': 1e-4,
                r'pinv_chol': 1e-5,
            }, 1e-11))
        
    def check_tree_decomp(self, conversion):
        for n in self.sizes:
            K = self.randsymmat(n)
            b = self.randmat(n)
            d1 = self.decompclass(K)
            q1 = d1.quad(b)
            d2 = conversion(d1)
            q2 = d2.quad(b)
            util.assert_equal(q1, q2)
    
    def test_flatten_decomp(self):
        def conversion(d):
            c, a = jax.tree_util.tree_flatten(d)
            return jax.tree_util.tree_unflatten(a, c)
        self.check_tree_decomp(conversion)
    
    def check_decomp_aux(self, op):
        def conversion(d):
            _, d2 = op(lambda x: (x, d), has_aux=True)(0.)
            return d2
        self.check_tree_decomp(conversion)
    
    def test_decomp_aux_jacfwd(self):
        self.check_decomp_aux(jax.jacfwd)
    
    def test_decomp_aux_jacrev(self):
        self.check_decomp_aux(jax.jacrev)
    
    def test_decomp_aux_grad(self):
        self.check_decomp_aux(jax.grad)
    
    # TODO test derivatives w.r.t. b, c
    # reverse gradient is broken in empbayes_fit, so I envisage that reverse
    # derivatives of quad w.r.t. b and c won't work due to the quad_autodiff
    # in the custom jvp

    def test_correlate_eye(self):
        self._woodbury_possign = True
        for n in self.sizes:
            K = self.randsymmat(n)
            dec = self.decompclass(K)
            A = dec.correlate(np.eye(dec.m))
            Q = A @ A.T
            util.assert_close_matrices(K, Q, rtol=self.clskey({
                r'pinv2_chol': 1e-4,
                r'pinv_chol': 1e-10,
                r'lobpcg': 1e-12,
            }, 1e-13))
    
    def test_double_correlate(self):
        self._woodbury_possign = True
        for n in self.sizes:
            K = self.randsymmat(n)
            d = self.decompclass(K)
            K2 = d.correlate(d.correlate(np.eye(n), transpose=True))
            util.assert_close_matrices(K, K2, atol=1e-15, rtol=self.clskey({
                r'pinv2_chol': 1e-3,
                r'pinv_chol': 1e-10,
            }, 1e-13))
    
    def test_correlate_transpose(self):
        self._woodbury_possign = True
        for n in self.sizes:
            K = self.randsymmat(n)
            AT = self.decompclass(K).correlate(np.eye(n), transpose=True)
            K2 = AT.T @ AT
            util.assert_close_matrices(K, K2, atol=1e-15, rtol=self.clskey({
                r'pinv_chol': 1e-9,
            }, 1e-13))

    def test_decorrelate_mat(self):
        self._woodbury_possign = True
        for n in self.sizes:
            K = self.randsymmat(n)
            b = self.randmat(n)
            x = self.decompclass(K).decorrelate(b)
            result = x.T @ x
            sol = self.decompclass(K).quad(b)
            util.assert_close_matrices(result, sol, rtol=self.clskey({
                r'pinv2_chol': 1e-6,
            }, 1e-13))
    
    def test_double_decorrelate(self):
        self._woodbury_possign = True
        for n in self.sizes:
            K = self.randsymmat(n)
            d = self.decompclass(K)
            Kinv = d.inv()
            Kinv2 = d.decorrelate(d.decorrelate(np.eye(n)), transpose=True)
            util.assert_close_matrices(Kinv, Kinv2, rtol=self.clskey({
                r'pinv2_chol': 1e-4,
                r'pinv_chol_lowrank': 1e-13,
            }, 1e-14))

class CompositeDecompTestBase(DecompTestBase):
    """ Abstract class to test decompositions which do not take as input
    the matrix itself """
    
    @property
    @abc.abstractmethod
    def compositeclass(self): pass
    
    @property
    @abc.abstractmethod
    def subdecompclass(self): pass
    
    @staticmethod
    def splitrank(n1, n2, rank):
        r = rng.integers(max(1, rank - n2), min(rank, n1 + 1)) if rank > 1 else 0
        assert 0 <= r <= n1 and 0 <= rank - r <= n2
        return r, rank - r

class ToeplitzBase(DecompTestBase):
    
    @staticmethod
    def singular_toeplitz(n, rank, ampl):
        """ produce an n x n real symmetric toeplitz matrix with given rank,
        ampl is an array of input values used deterministically in the
        construction """
        assert 0 <= rank <= n
        assert ampl.shape == (rank // 2 + rank % 2,)
        freqs = jnp.linspace(0, 1, 2 + n)[1 - rank % 2:1 + rank // 2]
        i = jnp.arange(n)[:, None]
        j = jnp.arange(n)[None, :]
        comps = jnp.cos(2 * jnp.pi * (i - j) * freqs[:, None, None]) * jnp.abs(ampl[:, None, None])
        return jnp.sum(comps, axis=0)

    def randsymmat(self, n):
        rank = self.rank(n)
        ampl = rng.gamma(2, 1/2, rank // 2 + rank % 2)
        return self.singular_toeplitz(n, rank, ampl)
    
    def mat(self, s, n):
        rank = self.rank(n)
        ampl = 1 + jnp.cos(s + jnp.arange(rank // 2 + rank % 2))
        return self.singular_toeplitz(n, rank, ampl)

class BlockDecompTestBase(CompositeDecompTestBase):
    """
    Abstract class for testing BlockDecomp.
    """
    
    def decompclass(self, K, **kw):
        if len(K) == 1:
            return self.subdecompclass(K, **kw)
        p = rng.integers(1, len(K))
        P = K[:p, :p]
        Q = K[:p, p:]
        S = K[p:, p:]
        Pdec = self.subdecompclass(P, **kw)
        subdec = lambda K, **kw: self.subdecompclass(K, **kw)
        return _linalg.Block(Pdec, S, Q, subdec, **kw)
    
class BlockDiagDecompTestBase(CompositeDecompTestBase):
    """
    Abstract class for testing BlockDiagDecomp.
    """
    
    _p = functools.cached_property(lambda self: {})
    def p(self, n):
        return self._p.setdefault(n, rng.integers(1, n) if n > 1 else 0)
    
    _subranks = functools.cached_property(lambda self: {})
    def subranks(self, n, rank):
        p = self.p(n)
        return self._subranks.setdefault(n, self.splitrank(p, n - p, rank))
    
    def randsymmat(self, n):
        rank = self.rank(n)
        p = self.p(n)
        r1, r2 = self.subranks(n, rank)
        assert 0 <= p <= n and 0 <= r1 <= p and 0 <= r2 <= n - p and r1 + r2 == rank
        K = np.zeros((n, n))
        if p > 0:
            K[:p, :p] = super().randsymmat(p, rank=r1)
        K[p:, p:] = super().randsymmat(n - p, rank=r2)
        return K
    
    def mat(self, s, n):
        rank = self.rank(n)
        p = self.p(n)
        r1, r2 = self.subranks(n, rank)
        A = super().mat(s, p, rank=r1)
        B = super().mat(s, n - p, rank=r2)
        return jlinalg.block_diag(A, B)
                
    def decompclass(self, K, **kw):
        if len(K) == 1:
            return self.subdecompclass(K, **kw)
        p = self.p(len(K))
        A = K[:p, :p]
        B = K[p:, p:]
        args = (self.subdecompclass(A, **kw), self.subdecompclass(B, **kw))
        return _linalg.BlockDiag(*args, **kw)

class SandwichTestBase(CompositeDecompTestBase):
    """
    Abstract base class to test Sandwich* decompositions
    """
    
    _inner = functools.cached_property(lambda self: {})
    def inner(self, n):
        return self._inner.setdefault(n, 1 + rng.integers(n))
            
    _rank = functools.cached_property(lambda self: {})
    def rank(self, n):
        return self._rank.setdefault(n, 1 + rng.integers(self.inner(n)) if self.lowrank else self.inner(n))
        # TODO guaranteed low rank
        
    _B = functools.cached_property(lambda self: {})
    def B(self, n):
        return self._B.setdefault(n, rng.standard_normal((n, self.rank(n))))
    
    _randA = functools.cached_property(lambda self: {})
    def randA(self, n):
        return self._randA.setdefault(n, super().randsymmat(self.inner(n), rank=self.rank(n)))
    
    def matA(self, s, n):
        return super().mat(s, self.inner(n), rank=self.rank(n))
    
    def randsymmat(self, n):
        A = self.randA(n)
        B = self.B(n)
        K = B @ A @ B.T
        util.assert_close_matrices(K, K.T, rtol=1e-15)
        self._afrom = 'randsymmat'
        return K
    
    def mat(self, s, n):
        A = self.matA(s, n)
        B = self.B(n)
        K = B @ A @ B.T
        self._afrom = 'mat'
        self._sparam = s
        return K
        
    def decompclass(self, K, **kw):
        if self._afrom == 'randsymmat':
            A = self.randA(len(K))
        elif self._afrom == 'mat':
            A = self.matA(self._sparam, len(K))
        B = self.B(len(K))
        A_decomp = self.subdecompclass(A, **kw)
        return self.compositeclass(A_decomp, B, **kw)
            
    def quad_or_solve(self, K, *args, **kw):
        sol, rank = super().quad_or_solve(K, *args, return_rank=True, **kw)
        assert rank == self.rank(len(K))
        return sol
    
class WoodburyTestBase(CompositeDecompTestBase):
    """
    Abstract base class to test Woodbury*
    """
    
    # TODO need a way to specify I want the low rank only on the inner matrix
    # to test woodbury2
    
    _inner = functools.cached_property(lambda self: {})
    def inner(self, n):
        return self._inner.setdefault(n, rng.integers(1, n + 1))
    
    _sign = functools.cached_property(lambda self: {})
    def sign(self, n):
        if getattr(self, '_woodbury_possign', False):
            return 1
        return self._sign.setdefault(n, -1 + 2 * rng.integers(2))
    
    _randM = functools.cached_property(lambda self: {})
    def randM(self, n):
        return self._randM.setdefault(n, super().randsymmat(n + self.inner(n)))
    
    def matM(self, s, n):
        return super().mat(s, n + self.inner(n))
        
    _subranks = functools.cached_property(lambda self: {})
    def subranks(self, n):
        return self._subranks.setdefault(n, self.splitrank(n, self.inner(n), self.rank(n + self.inner(n))))
    
    _proj = functools.cached_property(lambda self: {})
    def proj(self, n):
        u1 = randortho(n)
        u2 = randortho(self.inner(n))
        r1, r2 = self.subranks(n)
        w1 = np.zeros(len(u1))
        w2 = np.zeros(len(u2))
        w1[:r1] = 1
        w2[:r2] = 1
        p1 = (u1 * w1) @ u1.T
        p2 = (u2 * w2) @ u2.T
        return self._proj.setdefault(n, (p1, p2))
    
    def ABC(self, n, M):
        A = M[:n, :n]
        B = M[:n, n:]
        C = M[n:, n:]
        pa, pc = self.proj(n)
        A = pa @ A @ pa
        B = pa @ B @ pc
        C = pc @ C @ pc
        C = jnp.linalg.pinv(C) # such that A - BCB^T is p.s.d.
        return A, B, C
        
    def randsymmat(self, n):
        A, B, C = self.ABC(n, self.randM(n))
        K = A + self.sign(n) * B @ C @ B.T
        util.assert_close_matrices(K, K.T, rtol=1e-14)
        self._mfrom = 'randsymmat'
        return K
    
    def mat(self, s, n):
        A, B, C = self.ABC(n, self.matM(s, n))
        K = A + self.sign(n) * B @ C @ B.T
        self._mfrom = 'mat'
        self._sparam = s
        return K
    
    def decompclass(self, K, **kw):
        n = len(K)
        if self._mfrom == 'randsymmat':
            M = self.randM(n)
        elif self._mfrom == 'mat':
            M = self.matM(self._sparam, n)
        A, B, C = self.ABC(n, M)
        A_decomp = self.subdecompclass(A, **kw)
        C_decomp = self.subdecompclass(C, **kw)
        return self.compositeclass(A_decomp, B, C_decomp, self.subdecompclass, sign=self.sign(n), **kw)

class PinvTestBase(CompositeDecompTestBase):
    
    def decompclass(self, K, **kw):
        return self.compositeclass(K, self.subdecompclass, **kw)

class Pinv2TestBase(CompositeDecompTestBase):
    
    def decompclass(self, K, **kw):
        decompcls = lambda *args, **kw: self.subdecompclass(*args, epsrel=0, **kw)
        return self.compositeclass(K, decompcls, epsrel=1e-10, N=2, **kw)

class TestWoodbury2_EigCutFullRank(WoodburyTestBase): pass
class TestWoodbury_EigCutFullRank(WoodburyTestBase): pass

class TestSandwichSVD_EigCutFullRank(SandwichTestBase): pass
class TestSandwichQR_EigCutFullRank(SandwichTestBase): pass

class TestBlockDiag_EigCutFullRank(BlockDiagDecompTestBase): pass

class TestBlock_EigCutFullRank(BlockDecompTestBase): pass

class TestCholToeplitz(ToeplitzBase): pass

class TestChol(DecompTestBase): pass
class TestSVDCutLowRank(DecompTestBase): pass
class TestSVDCutFullRank(DecompTestBase): pass
class TestEigCutLowRank(DecompTestBase): pass
class TestEigCutFullRank(DecompTestBase): pass
        
class TestLanczos(DecompTestBase):
    def decompclass(self, K, **kw):
        return _linalg.Lanczos(K, rank=self.rank(len(K)), **kw)
    
class TestLOBPCG(DecompTestBase):
    def decompclass(self, K, **kw):
        return _linalg.LOBPCG(K, rank=self.rank(len(K)), **kw)

@mark.skip # Pinv and Pinv2 are crap, don't bother testing them for now
class TestPinv_Chol(PinvTestBase): pass
@mark.skip
class TestPinv2_Chol(Pinv2TestBase): pass

class TestSVDCutLowRank_LowRank(DecompTestBase): pass
class TestEigCutLowRank_LowRank(DecompTestBase): pass
class TestLanczos_LowRank(TestLanczos): pass
class TestLOBPCG_LowRank(TestLOBPCG): pass
@mark.skip
class TestPinv_Chol_LowRank(PinvTestBase): pass
@mark.skip
class TestPinv2_Chol_LowRank(Pinv2TestBase): pass
# class TestEigCutFullRank_LowRank(DecompTestBase): pass # fails as expected

@util.tryagain
def test_solve_triangular():
    for n in DecompTestBase.sizes:
        for ndim in range(4):
            for lower in [True, False]:
                tri = np.tril if lower else np.triu
                A = tri(rng.standard_normal((n, n)))
                diag = np.sqrt(np.sum(rng.standard_normal((n, 2)) ** 2, axis=-1) / 2)
                A[np.diag_indices(n)] = diag
                shape = rng.integers(1, 4, size=ndim)
                B = rng.standard_normal((*shape[:-1], n, *shape[-1:]))
                x = _linalg.solve_triangular(A, B, lower=lower)
                lhs = (A @ x).reshape(-1, B.shape[-1])
                util.assert_close_matrices(lhs, B.reshape(lhs.shape), rtol=1e-12)

@util.tryagain
def test_toeplitz_gershgorin():
    t = rng.standard_normal(100)
    m = linalg.toeplitz(t)
    b1 = _linalg._decomp._gershgorin_eigval_bound(m)
    b2 = _linalg._toeplitz.eigv_bound(t)
    util.assert_close_matrices(b2, b1, rtol=1e-15)

def check_toeplitz():
    mod = _linalg._toeplitz
    for n in [10, 2, 1]:
        x = np.linspace(0, 3, n)
        t = np.pi * np.exp(-1/2 * x ** 2)
        m = linalg.toeplitz(t)
    
        l1 = mod.chol(t)
        l2 = linalg.cholesky(m, lower=True)
        util.assert_close_matrices(l1, l2, rtol=1e-10)
    
        b = rng.standard_normal((len(t), 30))
        lb1 = mod.chol_matmul(t, b)
        lb2 = l2 @ b
        util.assert_close_matrices(lb1, lb2, rtol=1e-10)

        ld1 = mod.logdet(t)
        _, ld2 = np.linalg.slogdet(m)
        util.assert_close_matrices(ld1, ld2, rtol=1e-9)
    
        ilb1 = mod.chol_solve(t, b)
        ilb2 = linalg.solve_triangular(l2, b, lower=True)
        util.assert_close_matrices(ilb1, ilb2, rtol=1e-8)
    
        imb1 = mod.solve(t, b)
        imb2 = np.linalg.solve(m, b)
        util.assert_close_matrices(imb1, imb2, rtol=1e-8)

@util.tryagain
def test_toeplitz_nojit():
    with jax.disable_jit():
        check_toeplitz()

@util.tryagain
def test_toeplitz():
    check_toeplitz()

@util.tryagain
def test_toeplitz_chol_solve_numpy():
    shapes = [
        [(), ()],
        [(10,), (1,)],
        [(1, 2), (3, 1)],
        [(1, 4), (4,)],
        [(3,), (1, 3)],
    ]
    for tshape, bshape in shapes:
        for n in [0, 1, 2, 10]:
            x = np.linspace(0, 3, n)
            gamma = rng.uniform(0, 2, tshape + (1,))
            t = np.pi * np.exp(-1/2 * x ** gamma)
            m = np.empty(tshape + (n, n))
            for i in np.ndindex(*tshape):
                m[i] = linalg.toeplitz(t[i])
            l = np.linalg.cholesky(m)
            for shape in [(), (1,), (2,), (10,)]:
                if bshape and not shape:
                    continue
                b = rng.standard_normal((*bshape, n, *shape))
                ilb = _linalg._toeplitz.chol_solve_numpy(t, b, diageps=1e-12)
                lhs, rhs = np.broadcast_arrays(l @ ilb, b)
                lhs = lhs.reshape(-1 if lhs.size else 0, lhs.shape[-1])
                rhs = rhs.reshape(lhs.shape)
                util.assert_close_matrices(lhs, rhs, rtol=1e-7)
    with pytest.raises(np.linalg.LinAlgError):
        _linalg._toeplitz.chol_solve_numpy([-1], [1])
    with pytest.raises(np.linalg.LinAlgError):
        _linalg._toeplitz.chol_solve_numpy([1, 2], [1, 1])
    with pytest.raises(np.linalg.LinAlgError):
        _linalg._toeplitz.chol_solve_numpy([1, 0.5, 2], [1, 1, 1])

@mark.parametrize('mode', ['reduced', 'full'])
def test_rq(mode):
    for n in DecompTestBase.sizes:
        a = rng.standard_normal((n, 5))
        r, q = _linalg._decomp._rq(a, mode=mode)
        util.assert_close_matrices(r @ q, a, rtol=1e-15)
        util.assert_equal(np.tril(r), r)
        util.assert_close_matrices(  q @ q.T @   q,   q, rtol=1e-14)
        util.assert_close_matrices(q.T @   q @ q.T, q.T, rtol=1e-14)

@mark.parametrize('decomp', [
    _linalg.Chol,
    _linalg.EigCutLowRank,
    _linalg.SVDCutLowRank,
])
def test_degenerate(decomp):
    a = linalg.block_diag(np.eye(5), 0)
    d = decomp(a)
    util.assert_close_matrices(d.quad(a), a, rtol=1e-14)

def test_acausal_alg():
    with pytest.raises(ValueError):
        _linalg._seqalg.sequential_algorithm(2, [_linalg._seqalg.Stack(0)])

#### XFAILS ####
# keep last to avoid hiding them in wrappings

# TODO second derivatives completing but returning incorrect result
util.xfail(DecompTestBase, 'test_solve_matrix_hess_fwd_rev') # works only for TestSandwichSVDDiag
util.xfail(DecompTestBase, 'test_logdet_hess_fwd_rev') # since on quad it works, maybe I can solve it by using quad in the custom jvp
util.xfail(WoodburyTestBase, 'test_logdet_hess_fwd_fwd_stopg')
util.xfail(WoodburyTestBase, 'test_quad_matrix_matrix_hess_fwd_fwd_stopg')

# TODO fail on ubuntu-latest since jax 0.4.6 (???????), it's all the _da methods
# for Chol (completing but incorrect result, no inf/nan)
util.xfail(TestChol, 'test_solve_matrix_jac_fwd_da')
util.xfail(TestChol, 'test_solve_matrix_hess_da')
util.xfail(TestChol, 'test_quad_matrix_matrix_jac_fwd_da')
util.xfail(TestChol, 'test_quad_matrix_matrix_hess_da')
util.xfail(TestChol, 'test_logdet_hess_da')

# TODO linalg.sparse.eigsh/lobpcg are not implemented in jax
for testcls in [TestLanczos, TestLOBPCG]:
    for name, meth in inspect.getmembers(testcls, inspect.isfunction):
        if name.endswith('_da'):
            util.xfail(testcls, name)

# TODO these derivatives are a full nan, no idea, but it depends on the values,
# and on the sign, so they xpass at random. really no idea
util.xfail(TestWoodbury2_EigCutFullRank, 'test_solve_matrix_jac_fwd_da')
util.xfail(TestWoodbury2_EigCutFullRank, 'test_solve_matrix_hess_da')
util.xfail(TestWoodbury2_EigCutFullRank, 'test_quad_matrix_matrix_jac_fwd_da')
util.xfail(TestWoodbury2_EigCutFullRank, 'test_quad_matrix_matrix_hess_da')
util.xfail(TestWoodbury2_EigCutFullRank, 'test_logdet_hess_da')
util.xfail(TestBlock_EigCutFullRank, 'test_quad_vec_jac_fwd_jit')

# TODO fail with full nan
util.xfail(TestSVDCutLowRank_LowRank, 'test_solve_matrix_hess_da')
util.xfail(TestSVDCutLowRank_LowRank, 'test_quad_matrix_matrix_hess_da')
util.xfail(TestSVDCutLowRank_LowRank, 'test_logdet_hess_da')
util.xfail(TestEigCutLowRank_LowRank, 'test_solve_matrix_hess_da')
util.xfail(TestEigCutLowRank_LowRank, 'test_quad_matrix_matrix_hess_da')
util.xfail(TestEigCutLowRank_LowRank, 'test_logdet_hess_da')

# TODO sometimes very inaccurate. Maybe all these problems would go away if I
# tried to understand well what I'm doing with the parametric matrix to avoid
# exploding derivatives. Possibly the proposed unification of randsymmat() and
# mat() would do it.
util.xfail(TestWoodbury_EigCutFullRank, 'test_solve_matrix_jac_rev_matrix')
util.xfail(TestWoodbury_EigCutFullRank, 'test_solve_matrix_jac_rev_jit')
util.xfail(TestWoodbury_EigCutFullRank, 'test_solve_matrix_hess_fwd_fwd')
util.xfail(TestWoodbury_EigCutFullRank, 'test_quad_vec_jac_rev_jit')
util.xfail(TestWoodbury_EigCutFullRank, 'test_quad_vec_jac_rev')
util.xfail(TestWoodbury_EigCutFullRank, 'test_quad_vec_jac_fwd')
util.xfail(TestWoodbury_EigCutFullRank, 'test_logdet_jac_rev_jit')
util.xfail(TestWoodbury_EigCutFullRank, 'test_logdet_jac_rev')
util.xfail(TestWoodbury_EigCutFullRank, 'test_logdet_hess_fwd_fwd')

# TODO this fails because Pinv(2).logdet is not a pseudodeterminant.
util.xfail(TestPinv_Chol_LowRank, 'test_logdet')
util.xfail(TestPinv2_Chol_LowRank, 'test_logdet')

# TODO wildly inaccurate derivatives when the result is large, may xpass
util.xfail(TestPinv_Chol_LowRank, 'test_solve_vec_jac_rev')
util.xfail(TestPinv_Chol_LowRank, 'test_solve_vec_jac_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_solve_matrix_jac_rev')
util.xfail(TestPinv_Chol_LowRank, 'test_solve_matrix_jac_fwd_matrix')
util.xfail(TestPinv_Chol_LowRank, 'test_solve_matrix_jac_fwd_da')
util.xfail(TestPinv_Chol_LowRank, 'test_solve_matrix_jac_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_solve_matrix_hess_fwd_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_solve_matrix_hess_da')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_vec_jac_rev')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_vec_jac_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_matrix_matrix_jac_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_matrix_matrix_hess_fwd_rev')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_matrix_matrix_hess_fwd_fwd_stopg')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_matrix_matrix_hess_fwd_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_matrix_matrix_hess_da')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_matrix_jac_rev')
util.xfail(TestPinv_Chol_LowRank, 'test_quad_matrix_jac_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_logdet_jac_rev')
util.xfail(TestPinv_Chol_LowRank, 'test_logdet_jac_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_logdet_hess_fwd_fwd')
util.xfail(TestPinv_Chol_LowRank, 'test_logdet_hess_da')
util.xfail(TestPinv_Chol, 'test_solve_vec_jac_rev')
util.xfail(TestPinv_Chol, 'test_solve_vec_jac_fwd')
util.xfail(TestPinv_Chol, 'test_solve_matrix_jac_rev_matrix')
util.xfail(TestPinv_Chol, 'test_solve_matrix_jac_rev')
util.xfail(TestPinv_Chol, 'test_solve_matrix_jac_fwd_matrix')
util.xfail(TestPinv_Chol, 'test_solve_matrix_jac_fwd_da')
util.xfail(TestPinv_Chol, 'test_solve_matrix_jac_fwd')
util.xfail(TestPinv_Chol, 'test_solve_matrix_hess_fwd_fwd')
util.xfail(TestPinv_Chol, 'test_solve_matrix_hess_da')
util.xfail(TestPinv_Chol, 'test_quad_vec_jac_rev')
util.xfail(TestPinv_Chol, 'test_quad_vec_jac_fwd')
util.xfail(TestPinv_Chol, 'test_quad_matrix_matrix_jac_fwd_da')
util.xfail(TestPinv_Chol, 'test_quad_matrix_matrix_jac_fwd')
util.xfail(TestPinv_Chol, 'test_quad_matrix_matrix_hess_fwd_rev')
util.xfail(TestPinv_Chol, 'test_quad_matrix_matrix_hess_fwd_fwd_stopg')
util.xfail(TestPinv_Chol, 'test_quad_matrix_matrix_hess_fwd_fwd')
util.xfail(TestPinv_Chol, 'test_quad_matrix_matrix_hess_da')
util.xfail(TestPinv_Chol, 'test_quad_matrix_jac_rev')
util.xfail(TestPinv_Chol, 'test_quad_matrix_jac_fwd')
util.xfail(TestPinv_Chol, 'test_logdet_jac_rev')
util.xfail(TestPinv_Chol, 'test_logdet_jac_fwd')
util.xfail(TestPinv_Chol, 'test_logdet_hess_fwd_fwd_stopg')
util.xfail(TestPinv_Chol, 'test_logdet_hess_fwd_fwd')
util.xfail(TestPinv_Chol, 'test_logdet_hess_da')
util.xfail(TestPinv2_Chol_LowRank, 'test_solve_vec_jac_fwd')
util.xfail(TestPinv2_Chol_LowRank, 'test_solve_matrix_jac_rev_matrix')
util.xfail(TestPinv2_Chol_LowRank, 'test_solve_matrix_jac_rev')
util.xfail(TestPinv2_Chol_LowRank, 'test_solve_matrix_jac_fwd_matrix')
util.xfail(TestPinv2_Chol_LowRank, 'test_solve_matrix_jac_fwd_da')
util.xfail(TestPinv2_Chol_LowRank, 'test_solve_matrix_jac_fwd')
util.xfail(TestPinv2_Chol_LowRank, 'test_solve_matrix_hess_fwd_fwd')
util.xfail(TestPinv2_Chol_LowRank, 'test_quad_vec_jac_rev')
util.xfail(TestPinv2_Chol_LowRank, 'test_quad_matrix_matrix_jac_fwd')
util.xfail(TestPinv2_Chol_LowRank, 'test_quad_matrix_matrix_hess_fwd_rev')
util.xfail(TestPinv2_Chol_LowRank, 'test_quad_matrix_matrix_hess_fwd_fwd_stopg')
util.xfail(TestPinv2_Chol_LowRank, 'test_quad_matrix_matrix_hess_fwd_fwd')
util.xfail(TestPinv2_Chol_LowRank, 'test_quad_matrix_matrix_hess_da')
util.xfail(TestPinv2_Chol_LowRank, 'test_quad_matrix_jac_rev')
util.xfail(TestPinv2_Chol_LowRank, 'test_logdet_jac_rev')
util.xfail(TestPinv2_Chol_LowRank, 'test_logdet_hess_fwd_fwd_stopg')
util.xfail(TestPinv2_Chol_LowRank, 'test_logdet_hess_fwd_fwd')
util.xfail(TestPinv2_Chol_LowRank, 'test_logdet_hess_da')
util.xfail(TestPinv2_Chol, 'test_solve_vec_jac_rev')
util.xfail(TestPinv2_Chol, 'test_solve_vec_jac_fwd')
util.xfail(TestPinv2_Chol, 'test_solve_matrix_jac_rev_matrix')
util.xfail(TestPinv2_Chol, 'test_solve_matrix_jac_rev')
util.xfail(TestPinv2_Chol, 'test_solve_matrix_jac_fwd')
util.xfail(TestPinv2_Chol, 'test_solve_matrix_hess_fwd_fwd')
util.xfail(TestPinv2_Chol, 'test_solve_matrix_hess_da')
util.xfail(TestPinv2_Chol, 'test_quad_vec_jac_fwd')
util.xfail(TestPinv2_Chol, 'test_quad_matrix_matrix_jac_fwd_da')
util.xfail(TestPinv2_Chol, 'test_quad_matrix_matrix_jac_fwd')
util.xfail(TestPinv2_Chol, 'test_quad_matrix_matrix_hess_fwd_rev')
util.xfail(TestPinv2_Chol, 'test_quad_matrix_matrix_hess_fwd_fwd_stopg')
util.xfail(TestPinv2_Chol, 'test_quad_matrix_matrix_hess_fwd_fwd')
util.xfail(TestPinv2_Chol, 'test_quad_matrix_matrix_hess_da')
util.xfail(TestPinv2_Chol, 'test_quad_matrix_jac_rev')
util.xfail(TestPinv2_Chol, 'test_quad_matrix_jac_fwd')
util.xfail(TestPinv2_Chol, 'test_logdet_jac_rev')
util.xfail(TestPinv2_Chol, 'test_logdet_jac_fwd')
util.xfail(TestPinv2_Chol, 'test_logdet_hess_fwd_fwd_stopg')
util.xfail(TestPinv2_Chol, 'test_logdet_hess_fwd_fwd')
util.xfail(TestPinv2_Chol, 'test_logdet_hess_da')
util.xfail(TestPinv2_Chol_LowRank, 'test_logdet_jac_fwd')

# TODO Pinv2's correlate is a mess!
util.xfail(TestPinv2_Chol_LowRank, 'test_double_decorrelate')
util.xfail(TestPinv2_Chol_LowRank, 'test_double_correlate')
util.xfail(TestPinv2_Chol_LowRank, 'test_decorrelate_mat')
util.xfail(TestPinv2_Chol_LowRank, 'test_correlate_transpose')
util.xfail(TestPinv2_Chol_LowRank, 'test_correlate_eye')
util.xfail(TestPinv2_Chol, 'test_decorrelate_mat')
util.xfail(TestPinv2_Chol, 'test_correlate_transpose')

# TODO inaccurate jit (??)
util.xfail(TestPinv_Chol_LowRank, 'test_logdet_jit')
util.xfail(TestPinv_Chol, 'test_solve_vec_jac_fwd_jit')
util.xfail(TestPinv_Chol, 'test_solve_matrix_jac_fwd_jit')
util.xfail(TestPinv_Chol, 'test_quad_vec_jac_fwd_jit')

# TODO why?
# util.xfail(BlockDecompTestBase, 'test_logdet_hess_num')



# lsqfitgp/_special/_exp.py
#
# Copyright (c) 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import jax
from jax import numpy as jnp

@jax.custom_jvp
@jax.jit
def expm1x(x):
    """
    Compute accurately exp(x) - 1 - x = x^2/2 1F1(1, 3, x).
    """
    n = 10 if x.dtype == jnp.float32 else 17
    k = jnp.arange(2, n + 1)
    f = jnp.cumprod(k)
    coef = jnp.array(1, x.dtype) / f[::-1]
    smallx = x * x * jnp.polyval(coef, x, unroll=n)
    return jnp.where(jnp.abs(x) < 1, smallx, jnp.expm1(x) - x)
    
    # see also the GSL
    # https://www.gnu.org/software/gsl/doc/html/specfunc.html#relative-exponential-functions

@expm1x.defjvp
def _expm1x_jvp(p, t):
    x, = p
    xt, = t
    return expm1x(x), jnp.expm1(x) * xt



import numpy as np
from matplotlib import pyplot as plt

import lsqfitgp as lgp

fig, ax = plt.subplots(num='K', clear=True)

x = np.linspace(0, 1e-10, 1000)
nu = 0
for eps in 0.01 * np.arange(10)[::-1]:
    v = nu + eps
    y = lgp.Matern(nu=v)(0, x)
    ax.plot(x, y, label=f'$\\nu = {v}$')

ax.legend()

fig.show()



import functools

import jax
from jax import numpy as jnp
from jax.scipy import special as jspecial
from matplotlib import pyplot as plt

from lsqfitgp import _patch_jax

@functools.partial(jax.custom_jvp, nondiff_argnums=(0, 1, 2, 3))
def taylor(coefgen, args, n, m, x):
    c = coefgen(n, n + m, *args)
    k = jnp.arange(n, n + m)
    c = c * jnp.exp(jspecial.gammaln(1 + k) - jspecial.gammaln(1 + k - n))
    return jnp.polyval(c[::-1], x)

@taylor.defjvp
def taylor_jvp(coefgen, args, n, m, primals, tangents):
    x, = primals
    xt, = tangents
    return taylor(coefgen, args, n, m, x), taylor(coefgen, args, n + 1, m, x) * xt

def coefgen_jv(s, e, nu):
    k = jnp.arange(s, e)
    m = k // 2
    return jnp.where(k % 2, 0, (-1) ** m / jnp.exp(jspecial.gammaln(1 + m) + jspecial.gammaln(1 + m + nu)))

def jvmod(nu, x):
    return taylor(coefgen_jv, (nu,), 0, 10, x / 2)

def jvmod2(nu, x):
    return (x / 2) ** -nu * _patch_jax.jv(nu, x)

fig, ax = plt.subplots(num='taylor', clear=True)

def deriv(f, i, a):
    if i == 0:
        return f
    else:
        return deriv(jax.grad(f, a), i - 1, a)
        
def vec1(f):
    return jax.vmap(f, (None, 0), 0)

x = 0.5 * jnp.linspace(0, 0.01, 1000)
nu = 1.3

for n in range(5):
    f1 = deriv(jvmod2, n, 1)
    f2 = deriv(jvmod, n, 1)
    f1 = vec1(f1)
    f2 = vec1(f2)
    y1 = f1(nu, x)
    y2 = f2(nu, x)
    ax.plot(x, y1, color=f'C{n}', label=f'{n}')
    ax.plot(x, y2, color=f'C{n}', linestyle='--')

ax.set_ylim(-1, 1)

fig.show()



# lsqfitgp/examples/even.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

Split a function into even and odd parts.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

gp = lgp.GP(lgp.ExpQuad())
gp.addproclintransf(lambda f: lambda x: (f(x) + f(-x)) / 2, [gp.DefaultProcess], 'even')
gp.addproclintransf(lambda f: lambda x: (f(x) - f(-x)) / 2, [gp.DefaultProcess], 'odd')

x1, y1 = 1, 1
gp.addx(x1, 'even', proc='even')

x2, y2 = 1, -1
gp.addx(x2, 'odd', proc='odd')

xplot = np.linspace(-5, 5, 300)
gp.addx(xplot, 'function')
gp.addx(xplot, 'even part', proc='odd')
gp.addx(xplot, 'odd part', proc='even')

y = gp.predfromdata({'even': y1, 'odd': y2})

fig, ax = plt.subplots(num='even', clear=True)

labels = [label for label in y if len(label) > 4]
for j, sample in enumerate(gvar.raniter(y, 2, eps=1e-16)):
    for i, label in enumerate(labels):
        ax.plot(xplot, sample[label], color=f'C{i}', alpha=0.8, label=label if j == 0 else None)
ax.plot([x1, x2], [y1, y2], '.k')
ax.legend()
    
fig.show()



import numpy as np
from matplotlib import pyplot as plt
import mpmath

import lsqfitgp as lgp

zeta = np.vectorize(lambda *args: float(mpmath.zeta(*args)))

fig, ax = plt.subplots(num='zeta2', clear=True)

a = np.linspace(0, 1, 1000)
for s in [-0.5]:
    z = lgp._patch_jax.hurwitz_zeta(s, a)
    zsure = zeta(s, a)
    ax.plot(a, np.abs(z - zsure) / np.max(np.abs(zsure)), label=f's={s:.2g}')

ax.set_yscale('log')
ax.set_xlabel('a')
ax.set_title('$|\\zeta_\\mathrm{impl}(s,a) - \\zeta(s, a)|\\,/\\,\\max_a|\\zeta(s, a)|$')
ax.axhline(np.finfo(float).eps, linestyle='--', color='black', label='1 ULP')
ax.axhline(100 * np.finfo(float).eps, linestyle='--', color='black', label='100 ULP')
ax.legend()

fig.show()



# lsqfitgp/tests/test_copula.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" Test the copula module """

import sys
import functools

import numpy as np
import gvar
import jax
import pytest

sys.path.insert(0, '.')
import lsqfitgp as lgp

class JaxConfig:
    
    def __init__(self, **opts):
        self.opts = opts
    
    def __enter__(self):
        self.prev = {k: jax.config.read(k) for k in self.opts}
        for k, v in self.opts.items():
            jax.config.update(k, v)
    
    def __exit__(self, *_):
        for k, v in self.prev.items():
            jax.config.update(k, v)

class CopulaBaseTest:
    
    name = 'lgp.copula.estdistr'
    
    def __init_subclass__(cls, params):
        assert cls.__name__.startswith('Test')
        cls.copcls = getattr(lgp.copula, cls.__name__[4:].lower())
        cls.params = params
    
    # don't make this a staticmethod because it wouldn't be callable at class
    # definition time in old python versions
    def clean(meth):
        @functools.wraps(meth)
        def newmeth(self, *args, **kw):
            if gvar.BufferDict.has_distribution(self.name):
                gvar.BufferDict.del_distribution(self.name)
            return meth(self, *args, **kw)
        return newmeth
    
    @clean
    def test_invfcn(self):
        self.copcls(self.name, *self.params)
        invfcn = gvar.BufferDict.invfcn[self.name]
        x = gvar.gvar(0.1, 1.2)
        y = invfcn(x)
        deriv = jax.grad(invfcn)(x.mean)
        ymean = invfcn(x.mean)
        ysdev = x.sdev * deriv
        np.testing.assert_allclose(y.mean, ymean, rtol=1e-6)
        np.testing.assert_allclose(y.sdev, ysdev, rtol=1e-6)
    
    @clean
    def test_overwrite(self):
        gvar.BufferDict.add_distribution(self.name, gvar.exp)
        with pytest.raises(ValueError):
            self.copcls(self.name, *self.params)
        gvar.BufferDict.del_distribution(self.name)
        self.copcls(self.name, *self.params)
        self.copcls(self.name, *self.params)
        with pytest.raises(ValueError):
            self.copcls(self.name, *(np.nan,) * len(self.params))
    
    @clean
    def test_bufferdict_gvar(self):
        key = f'{self.name}(x)'
        b = gvar.BufferDict({
            key: self.copcls(self.name, *self.params)
        })
        x = b['x']
        xmean = self.copcls.invfcn(b[key].mean, *self.params)
        xsdev = jax.grad(self.copcls.invfcn)(b[key].mean, *self.params)
        np.testing.assert_allclose(x.mean, xmean, rtol=1e-6)
        np.testing.assert_allclose(x.sdev, xsdev, rtol=1e-6)

    @clean
    def test_bufferdict(self):
        key = f'{self.name}(x)'
        self.copcls(self.name, *self.params)
        b = gvar.BufferDict({
            key: 0
        })
        x = b['x']
        x2 = self.copcls.invfcn(b[key], *self.params)
        np.testing.assert_allclose(x, x2, rtol=1e-6)

class TestInvGamma(CopulaBaseTest, params=(1.2, 2.3)): pass
class TestBeta(CopulaBaseTest, params=(1.2, 2.3)): pass

def test_invgamma_divergence():
    y = lgp.copula.invgamma.invfcn(10., 1, 1)
    assert np.isfinite(y)

def test_invgamma_zero():
    params = (1.2, 2.3)
    y = lgp.copula.invgamma.invfcn(-100, *params)
    assert y > 0
    with JaxConfig(jax_enable_x64=False):
        y1 = lgp.copula.invgamma.invfcn(-12 * (1 + np.finfo(np.float32).eps), *params)
        y2 = lgp.copula.invgamma.invfcn(-12 * (1 - np.finfo(np.float32).eps), *params)
        assert y1 > 0 and y2 > 0
        np.testing.assert_allclose(y1, y2, atol=0, rtol=2e-4)
    with JaxConfig(jax_enable_x64=True):
        y1 = lgp.copula.invgamma.invfcn(-37 * (1 + np.finfo(np.float64).eps), *params)
        y2 = lgp.copula.invgamma.invfcn(-37 * (1 - np.finfo(np.float64).eps), *params)
        assert y1 > 0 and y2 > 0
        np.testing.assert_allclose(y1, y2, atol=0, rtol=1e-5)

    # TODO improve the accuracy



# lsqfitgp/tests/test_deriv.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys

import numpy as np
from jax import numpy as jnp
import gvar
from scipy import stats
import pytest

sys.path.insert(0, '.')
import lsqfitgp as lgp

import util

def test_manyargs():
    with pytest.raises(ValueError):
        lgp.Deriv(1, 2)

def test_alienargs():
    with pytest.raises(TypeError):
        lgp.Deriv((None,))

def test_manyintegers():
    with pytest.raises(ValueError):
        lgp.Deriv((1, 2))

def test_alienarg():
    with pytest.raises(TypeError):
        lgp.Deriv(object)

def test_orphan():
    with pytest.raises(ValueError):
        lgp.Deriv((1,))
    with pytest.raises(ValueError):
        lgp.Deriv(('ciao', 1))

def test_length():
    assert len(lgp.Deriv([1, 'ciao', 2, 'pippo'])) == 2

def test_compare():
    assert not lgp.Deriv() == 'cippa'

def test_repr():
    assert repr(lgp.Deriv()) == '{}'



# lsqfitgp/tests/test_bart.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys

import numpy as np
import pytest
from pytest import mark

sys.path.insert(0, '.')
import lsqfitgp as lgp

import util

gen = np.random.default_rng(202207191826)

plist = [1, 5]
smark = mark.parametrize('sb,sbw,sa,w', sum([[
    (*gen.integers(0, 4, (3, p)), gen.integers(1, 10, p)),
    (*np.zeros((3, p), int), gen.integers(1, 10, p)),
    (np.zeros(p, int), np.pad([1], (0, p - 1)), np.zeros(p, int), gen.integers(1, 10, p)),
    (gen.integers(0, 10, p), (np.arange(p) == gen.integers(p)).astype(int), gen.integers(0, 10, p), gen.integers(1, 10, p)),
] for p in plist], []))
amark = mark.parametrize('a', [
    pytest.param(np.array([0, 1])[:, None], id='a01'),
    pytest.param(np.linspace(0.01, 0.99, 7)[:, None], id='aother'),
])
bmark = mark.parametrize('b', [
    pytest.param(np.array([0, np.inf]), id='b0inf'),
    pytest.param(np.linspace(1, 10, 10), id='bother'),
])
umark = mark.parametrize('u', [np.array([0, 1])[:, None, None]])
mdmark = mark.parametrize('md', range(5))

@mdmark
@bmark
@amark
@smark
def test_lower_lt_upper(sb, sbw, sa, w, a, b, md):
    """ 0 <= lower <= interp/stricter upper <= upper <= 1 """
    lw = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=0, maxd=md, weights=w)
    au = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma='auto' if 1 <= md <= 3 else 0, maxd=md, weights=w)
    vg = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=1, maxd=md * 2, reset=[md], weights=w)
    up = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=1, maxd=md, weights=w)
    np.testing.assert_array_max_ulp(np.zeros_like(lw), np.minimum(0, lw))
    np.testing.assert_array_max_ulp(lw, np.minimum(lw, au))
    np.testing.assert_array_max_ulp(lw, np.minimum(lw, vg))
    np.testing.assert_array_max_ulp(au, np.minimum(au, up))
    np.testing.assert_array_max_ulp(vg, np.minimum(vg, up))
    np.testing.assert_array_max_ulp(up, np.minimum(up, 1))

@bmark
@amark
@smark
def test_lower_upper_incr_maxd(sb, sbw, sa, w, a, b):
    """
    lower/upper increases/decreases as maxd is increased
    """
    for md in range(4):
        lw = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=0, maxd=md, weights=w)
        up = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=1, maxd=md, weights=w)
        if md:
            np.testing.assert_array_max_ulp(plw, np.minimum(lw, plw), 2)
            np.testing.assert_array_max_ulp(pup, np.maximum(up, pup), 2)
        plw = lw
        pup = up

@mdmark
@umark
@bmark
@amark
@smark
def test_incr_beta(sb, sbw, sa, w, a, b, u, md):
    """increases as beta is increased"""
    c = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, weights=w)
    ci = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b + 1e-3, gamma=u, maxd=md, weights=w)
    np.testing.assert_array_max_ulp(ci, np.maximum(c, ci))

@mdmark
@umark
@bmark
@amark
@smark
def test_incr_alpha(sb, sbw, sa, w, a, b, u, md):
    """decreases as alpha is increased"""
    da = 1e-3
    a = np.minimum(1 - da, a)
    c = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, weights=w)
    ci = lgp.BART.correlation(sb, sbw, sa, alpha=a + da, beta=b, gamma=u, maxd=md, weights=w)
    np.testing.assert_array_max_ulp(ci, np.minimum(c, ci))

@mdmark
@umark
@bmark
@amark
@mark.parametrize('sb,sbw,sa,w', sum([[
    # n^0 = 0
    (gen.integers(0, 10, p), np.zeros(p, int), gen.integers(0, 10, p), gen.integers(1, 10, p)),
    (*np.zeros((3, p), int), gen.integers(1, 10, p)),
] for p in plist], []) + sum([[ 
    # w = 0
    (*gen.integers(0, 4, (3, p)), np.zeros(p)),
] for p in plist], []) + sum([[
    # wi = 0 or ni = 0
    np.concatenate([gen.integers(0, 4, (3, p)), gen.integers(1, 10, (1, p))]) *
    (gen.integers(0, 2, p).astype(bool) ^ np.array([0, 0, 0, 1], bool)[:, None]),
] for p in plist], []) + [
    # p = 0
    (*np.empty((3, 0), int), np.empty(0)),
])
def test_corr_1(sb, sbw, sa, w, a, b, u, md):
    """
    correlation = 1 if:
        - n^0 = 0
        - w = 0
        - wi = 0 or ni = 0
        - p = 0
    """
    c = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, weights=w)
    np.testing.assert_array_max_ulp(c, np.ones_like(c))

@mdmark
@umark
@bmark
@amark
@smark
def test_swap_ab(sb, sbw, sa, w, a, b, u, md):
    """invariant under swapping of nplus and nminus"""
    swap = gen.integers(0, 2, sb.size)
    s1 = np.where(swap, sa, sb)
    s2 = np.where(swap, sb, sa)
    c = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, weights=w)
    cs = lgp.BART.correlation(s1, sbw, s2, alpha=a, beta=b, gamma=u, maxd=md, weights=w)
    np.testing.assert_array_max_ulp(c, cs, 16)

@mdmark
@umark
@bmark
@amark
@smark
def test_perm_dims(sb, sbw, sa, w, a, b, u, md):
    """invariant under reordering of the dimensions"""
    perm = gen.permutation(sb.size)
    c = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, weights=w)
    cp = lgp.BART.correlation(sb[perm], sbw[perm], sa[perm], alpha=a, beta=b, gamma=u, maxd=md, weights=w[perm])
    np.testing.assert_array_max_ulp(c, cp, 64)

@mdmark
@umark
@bmark
@amark
@mark.parametrize('sb,sbw,sa,w', sum([[
    (*gen.integers(0, 10, (3, p)), gen.integers(1, 10, p)),
] for p in plist], []))
def test_incr_n0(sb, sbw, sa, w, a, b, u, md):
    """correlation decreases as n0 increases at fixed ntot"""
    c = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, weights=w)

    ntot = sb + sbw + sa
    which = gen.permuted(np.eye(sb.size)[0]).astype(bool)
    dn = np.where(which & (sb + sa), 1, 0)
    lr = gen.integers(0, 2, sb.size).astype(bool)
    lr ^= lr & ~sb.astype(bool) | ~lr & ~sa.astype(bool)
    # lr sb sa
    # 0  *  0   1
    # 0  *  1   0
    # 1  0  *   0
    # 1  1  *   1
    sb -= np.where(lr, dn, 0)
    sbw += dn
    sa -= np.where(~lr, dn, 0)
    assert np.all(ntot == sb + sbw + sa)
    assert np.all(sb >= 0) and np.all(sa >= 0)

    ci = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, weights=w)
    np.testing.assert_array_max_ulp(ci, np.minimum(c, ci))

def values(mark):
    vals = mark.args[1]
    return sum((getattr(v, 'values', (v,)) for v in vals), ())

@mark.parametrize('sb,sbw,sa,w,a,b,md', [
    # n^0 = 0
    (gen.integers(0, 10, p), np.zeros(p, int), gen.integers(0, 10, p), gen.integers(1, 10, p), a, b, d)
    for p in plist
    for a in values(amark)
    for b in values(bmark)
    for d in range(0, mdmark.args[1].stop)
] + [
    # alpha = 0
    (gen.integers(0, 10, p), gen.integers(1, 10, p), gen.integers(0, 10, p), gen.integers(1, 10, p), 0, b, d)
    for p in plist
    for b in values(bmark)
    for d in range(0, mdmark.args[1].stop)
] + [
    # beta = inf
    (gen.integers(0, 10, p), gen.integers(1, 10, p), gen.integers(0, 10, p), gen.integers(1, 10, p), a, np.inf, d)
    for p in plist
    for a in values(amark)
    for d in range(1, mdmark.args[1].stop)
])
def test_lower_eq_upper(sb, sbw, sa, w, a, b, md):
    """
    upper = lower in cases where the solution is exact
     - beta = inf if maxd > 0
     - alpha = 0
     - n^0 = 0
    """
    lw = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=0, maxd=md, weights=w)
    up = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=1, maxd=md, weights=w)
    np.testing.assert_array_max_ulp(lw, up)

@mark.parametrize('md', range(4))
@umark
@bmark
@amark
@smark
def test_no_shortcuts(sb, sbw, sa, w, a, b, u, md):
    """check the result is the same if no recursions are avoided"""
    c = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, weights=w)
    cd = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, debug=True, weights=w)
    np.testing.assert_array_max_ulp(c, cd, 32)

@mdmark
@umark
@bmark
@amark
@smark
def test_nzero(sb, sbw, sa, w, a, b, u, md):
    """adding zeros in sb, sbw, sa does not change the result"""
    kw = dict(alpha=a, beta=b, gamma=u, maxd=md)
    c = lgp.BART.correlation(sb, sbw, sa, weights=w, **kw)
    z = lambda x, f=0, p=4: np.concatenate([x, np.full(p, f)])
    c0 = lgp.BART.correlation(z(sb), z(sbw), z(sa), weights=z(w, 1), **kw)
    np.testing.assert_array_max_ulp(c, c0, 4)

@mdmark
@umark
@bmark
@amark
@smark
def test_wzero(sb, sbw, sa, w, a, b, u, md):
    """adding zeros to weights with random n does not change the result"""
    kw = dict(alpha=a, beta=b, gamma=u, maxd=md)
    c = lgp.BART.correlation(sb, sbw, sa, weights=w, **kw)
    z = lambda x, f=5, p=4: np.concatenate([x, gen.integers(f + 1, size=p)])
    c0 = lgp.BART.correlation(z(sb), z(sbw), z(sa), weights=z(w, 0), **kw)
    np.testing.assert_array_max_ulp(c, c0, 4)

def test_structured():
    X = np.arange(10 * 2.).reshape(1, -1, 2).view('d,d')
    splits = lgp.BART.splits_from_coord(X)
    cov = lgp.BART(splits=splits)(X, X.T)

def test_duplicates():
    x = np.repeat(np.arange(10 * 2.).reshape(-1, 2), 2, axis=0).view('d,d').squeeze(axis=-1)
    length, splits = lgp.BART.splits_from_coord(x)
    assert np.all(length == 9)

def test_integer():
    X = np.arange(10 * 2).reshape(1, -1, 2)
    X1 = X.astype('d').view('d,d').squeeze(-1)
    X2 = X.astype('l').view('l,l').squeeze(-1)
    kernel1 = lgp.BART(splits=lgp.BART.splits_from_coord(X1))
    kernel2 = lgp.BART(splits=lgp.BART.splits_from_coord(X2))
    v1 = kernel1(X1, X1.T)
    v2 = kernel2(X1, X1.T)
    util.assert_equal(v1, v2)

def test_pnt():
    sss = [1, 2], [3, 1], [5, 8]
    alpha = 0.9
    beta = 1.6
    c1 = lgp.BART.correlation(*sss, alpha=alpha, beta=beta, maxd=2)
    c2 = lgp.BART.correlation(*sss, alpha=None, beta=None, pnt=[alpha, alpha/2 ** beta, alpha/3 ** beta])
    util.assert_allclose(c1, c2)

def test_wrong_gamma():
    with pytest.raises(KeyError):
        lgp.BART.correlation([0], [0], [0], gamma='ciao')

def test_intercept():
    sss = [1, 2], [3, 1], [5, 8]
    alpha = 0.9
    c1 = lgp.BART.correlation(*sss, alpha=alpha)
    c2 = lgp.BART.correlation(*sss, alpha=alpha, intercept=False)
    c2 = c2 * alpha + (1 - alpha)
    util.assert_allclose(c1, c2, rtol=1e-15)

def test_splits_1d():
    l = [2]
    s = np.array([-1/2, 1/2])
    x = np.array([-1, 0, 1])[:, None]
    k1 = lgp.BART(splits=(l, s))
    k2 = lgp.BART(splits=(l, s[:, None]))
    v1 = k1(x, x.T)
    v2 = k1(x, x.T)
    util.assert_equal(v1, v2)

def test_f32():
    sss = [1, 2], [3, 1], [5, 8]
    c = lgp.BART.correlation(*sss, pnt=np.array([0.9, 0.4, 0.3], 'f'))
    assert c.dtype == 'f'

def test_i32():
    """ test that passing 32 bit integers does not result in 32 bit floating
    point calculations, which would happen because jax casts int32 to float32
    """
    sss = [1, 2], [3, 1], [5, 8]
    sss = [np.array(s, np.int32) for s in sss]
    pnt = np.array([0.9, 0.4, 0.3], np.float64)
    c1 = lgp.BART.correlation(*sss, pnt=pnt)
    assert c1.dtype == np.float64
    sss = [np.array(s, np.int64) for s in sss]
    c2 = lgp.BART.correlation(*sss, pnt=pnt)
    assert c2.dtype == np.float64
    np.testing.assert_array_max_ulp(c1, c2, 0)

@mark.parametrize('md,reset', [
    (0, None),
    (1, None),
    (2, None), (2, [1]),
    (3, None), (3, [1, 2]),
    (4, None), (4, [2]), (4, [1, 2, 3]),
])
@umark
@bmark
@amark
@smark
def test_altinput(sb, sbw, sa, w, a, b, u, md, reset):
    """ two alternative implementations give the same result """
    c1 = lgp.BART.correlation(sb, sbw, sa, alpha=a, beta=b, gamma=u, maxd=md, reset=reset, weights=w)
    n = sb + sbw + sa
    ix = sb
    iy = sb + sbw
    ix, iy = np.broadcast_arrays(ix, iy)
    swap = gen.integers(0, 2, size=ix.shape, dtype=bool)
    ix, iy = np.where(swap, iy, ix), np.where(swap, ix, iy)
    c2 = lgp.BART.correlation(n, ix, iy, alpha=a, beta=b, gamma=u, maxd=md, reset=reset, weights=w, altinput=True)
    util.assert_allclose(c1, c2, rtol=1e-16, atol=2e-15)

def test_index_input():
    """ passing coordinates or directly indices gives the same result """
    n, p = 100, 10
    asstruct = lambda x: np.asarray(x).view([('f0', x.dtype, p)]).squeeze(-1)
    X = asstruct(gen.standard_normal((n, p)))
    splits = lgp.BART.splits_from_coord(X)
    x, y = asstruct(gen.standard_normal((2, n, p)))
    x, y = np.ix_(x, y)
    ix = asstruct(lgp.BART.indices_from_coord(x, splits))
    iy = asstruct(lgp.BART.indices_from_coord(y, splits))
    c1 = lgp.BART(splits=splits)(x, y)
    c2 = lgp.BART(splits=splits, indices=True)(ix, iy)
    np.testing.assert_array_max_ulp(c1, c2, 0)

# TODO
# - increases at fixed n0 and ntot if the difference between nminus and nplus
#   decreases (not completely sure)
# - test gamma='auto' gives 0 for beta=0, alpha=1
# - test of equality with precomputed values sampled with qmc
# - duplicated entries in reset have no effect
# - continuity w.r.t. weight in zero, unless all differring covariates disappear
#   and the points collapse (which should still be continuous if gamma=1)
# - reset
# - splitting points are checked only if indices=False



from scipy import linalg
import numpy as np

gen = np.random.default_rng(20230130003)

n = 10
rank = 5

freqs = np.linspace(0, 1, 2 + n)[1 - rank % 2:1 + rank // 2]
ampl = gen.gamma(1, 1, rank // 2 + rank % 2)
i = np.arange(n)[:, None]
j = np.arange(n)[None, :]
T_comps = np.cos(2 * np.pi * (i - j) * freqs[:, None, None]) * ampl[:, None, None]
T = np.sum(T_comps, axis=0)

assert np.allclose(T, T.T.conj())
assert np.allclose(T, linalg.toeplitz(T[:, 0]))
w = linalg.eigvalsh(T)
assert np.all(w[n - rank:] >= 0)
Tp, r = linalg.pinvh(T, return_rank=True)
assert r == rank



import gvar
import numpy as np

rng = np.random.default_rng(0)
n = 100
m = 101
a = rng.standard_normal((n + m, n + m))
cov = a.T @ a
c1 = cov[:n, :n]
c2 = cov[n:, n:]
c12 = cov[:n, n:]

x1 = gvar.gvar(np.zeros(n), c1, fast=True)
indices = np.arange(n)
rng.shuffle(indices)
x2 = gvar.gvar(np.zeros(m), c2, x1[indices], c12[indices], fast=True)
x = np.concatenate([x1, x2])

xcov = gvar.evalcov(x)
np.testing.assert_equal(xcov, cov)

y = gvar.gvar(np.zeros(n + m), cov, [], np.empty((0, n + m)), fast=True)
ycov = gvar.evalcov(y)
np.testing.assert_equal(ycov, cov)



import functools

import jax
from jax import numpy as jnp
from jax.scipy import special as jspecial
from matplotlib import pyplot as plt

@functools.partial(jax.custom_jvp, nondiff_argnums=(0, 1))
def sincder(n, m, x):
    start = (n + 1) // 2
    k = jnp.arange(start, start + m)
    sign = jnp.where(k % 2, -1, 1)
    c = sign * jnp.exp(jspecial.gammaln(1 + 2 * k) - jspecial.gammaln(1 + 1 + 2 * k) - jspecial.gammaln(1 + 2 * k - n))
    coeff = jnp.zeros(2 * len(c) - 1 + n % 2).at[n % 2::2].set(c)
    return jnp.polyval(coeff[::-1], x)

@sincder.defjvp
def sincder_jvp(n, m, primals, tangents):
    x, = primals
    xt, = tangents
    return sincder(n, m, x), sincder(n + 1, m, x) * xt

fig, ax = plt.subplots(num='sincder2', clear=True)

x = 0.5 * jnp.linspace(-1, 1, 1000)

for n in range(5):
    f = jnp.sinc
    for i in range(n):
        f = jax.grad(f)
    f = jax.vmap(f)
    y1 = f(x)
    y2 = jnp.pi ** n * sincder(n, 2, jnp.pi * x)
    ax.plot(x, y1, color=f'C{n}', label=f'{n}')
    ax.plot(x, y2, color=f'C{n}', linestyle='--')

fig.show()



import sys
import functools
import warnings

from jax import tree_util
import numpy as np
from jax import numpy as jnp
import gvar
import pytest
from scipy import linalg

sys.path.insert(0, '.')
import lsqfitgp as lgp

def jaxtonumpy(x):
    """
    Recursively convert jax arrays in x to numpy arrays.
    """
    children, meta = tree_util.tree_flatten(x)
    children = (np.array(x) if isinstance(x, jnp.ndarray) else x for x in children)
    return tree_util.tree_unflatten(meta, children)

def assert_equal(*args):
    """
    Version of assert_equal that works with jax arrays and StructuredArray, and
    which fixes numpy issue #21739
    """
    assert_array_equal(*jaxtonumpy(args))

def assert_array_equal(*args):
    args = [np.array(a) if isinstance(a, lgp.StructuredArray) else a for a in args]
    a = args[0]
    if isinstance(a, np.ndarray) and a.size == 0 and a.dtype.names:
        assert all(b.dtype == a.dtype for b in args)
        assert all(b.shape == a.shape for b in args)
    elif isinstance(a, np.ndarray) and a.dtype.names:
        # old versions of numpy force structured dtypes to be equal when
        # comparing, instead of casting
        assert all(b.dtype.names == a.dtype.names for b in args)
        for n in a.dtype.names:
            assert_array_equal(*(x[n] for x in args))
    elif isinstance(a, np.ndarray) and all(np.issubdtype(b.dtype, np.character) for b in args):
        # old versions of numpy do not compare as equal strings with different
        # allocated space
        basetype = a.dtype.type
        assert all(b.dtype.type == basetype for b in args) # check it's the same kind of string type
        basesize = np.dtype((basetype, 1)).itemsize
        maxsize = max(b.dtype.itemsize // basesize for b in args)
        newtype = np.dtype((basetype, maxsize))
        args = (b.astype(newtype) for b in args)
        np.testing.assert_equal(*args)
    else:
        np.testing.assert_equal(*args)

def mark(cls, meth, mark):
    """
    Function to mark a test method.
    """
    impl = getattr(cls, meth)
    if not meth.startswith('test_'):
        warnings.warn(f'method {cls.__name__}.{meth} not prefixed with test_')
    marker = getattr(pytest.mark, mark)
    @marker
    @functools.wraps(impl) # `wraps` needed because pytest uses the method name
    def newimpl(self):
        # wrap because otherwise the superclass method would be marked too,
        # but set the mark temporarily to allow introspection
        marker(impl)
        try:
            impl(self)
        finally:
            impl.pytestmark.pop()
    setattr(cls, meth, newimpl)

def xfail(cls, meth):
    mark(cls, meth, 'xfail')

def skip(cls, meth):
    mark(cls, meth, 'skip')

def tryagain(fun, rep=2, method=False):
    """
    Decorates test `fun` to make it try again in case of failure. Inhibited
    by xfails.
    """
    meta = {}
    @functools.wraps(fun)
    def newfun(*args, _meta=meta, **kw):

        job = lambda: fun(*args, **kw)

        marks = getattr(_meta['newfun'], 'pytestmark', [])
        if any(m.name == 'xfail' for m in marks):
            return job()

        if method:
            self = args[0]
            name = f'{self.__class__.__name__}.{fun.__name__}'
        else:
            name = fun.__name__

        for i in range(rep):
            try:
                x = job()
                if i > 0:
                    warnings.warn(f'Test {name} failed {i} times with exception {exc.__class__.__name__}: ' + ", ".join(exc.args))
                return x
            except Exception as e:
                exc = e

        # if rep > 1:
        #     warnings.warn(f'Test {name} failed {rep} times')
        raise exc
    
    if hasattr(fun, 'pytestmark'):
        newfun.pytestmark = fun.pytestmark
    meta['newfun'] = newfun
    return newfun

def assert_close_matrices(actual, desired, *, rtol=0, atol=0):
    actual = np.asarray(actual)
    desired = np.asarray(desired)
    assert actual.shape == desired.shape
    if actual.size == 0:
        return
    actual = np.atleast_1d(actual)
    desired = np.atleast_1d(desired)
    
    dnorm = linalg.norm(desired, 2)
    adnorm = linalg.norm(actual - desired, 2)
    ratio = adnorm / dnorm if dnorm else np.nan
    msg = f"""\
matrices actual and desired are not close in 2-norm
norm(desired) = {dnorm:.2g}
norm(actual - desired) = {adnorm:.2g}  (atol = {atol:.2g})
ratio = {ratio:.2g}  (rtol = {rtol:.2g})"""
    assert adnorm <= atol + rtol * dnorm, msg

def _assert_similar_gvars(g, h, rtol, atol):
    assert_allclose(gvar.mean(g), gvar.mean(h), rtol=rtol, atol=atol)
    g = np.reshape(g, -1)
    h = np.reshape(h, -1)
    assert_close_matrices(gvar.evalcov(g), gvar.evalcov(h), rtol=rtol, atol=atol)

def assert_similar_gvars(*gs, rtol=0, atol=0):
    if gs:
        for g in gs[1:]:
            _assert_similar_gvars(g, gs[0], rtol, atol)

def assert_same_gvars(g, h, *, atol=0):
    z = g - h
    z = np.reshape(z, -1)
    assert_allclose(gvar.mean(z), np.zeros(z.shape), rtol=0, atol=atol)
    assert_close_matrices(gvar.evalcov(z), np.zeros(2 * z.shape), rtol=0, atol=atol)

def assert_close_decomps(actual, desired, *, rtol=0, atol=0):
    assert actual.n == desired.n
    assert_close_matrices(desired.solve(actual.matrix()), np.eye(desired.n), rtol=rtol, atol=atol)
    assert_close_matrices(actual.solve(desired.matrix()), np.eye(desired.n), rtol=rtol, atol=atol)

def assert_allclose(actual, desired, *, rtol=0, atol=0, equal_nan=False, **kw):
    """ change the default arguments of np.testing.assert_allclose """
    np.testing.assert_allclose(actual, desired, rtol=rtol, atol=atol, equal_nan=equal_nan, **kw)



# lsqfitgp/_kernels/_randomwalk.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import jax
from jax import numpy as jnp

from .. import _patch_jax
from .._Kernel import kernel, stationarykernel

@kernel(forcekron=True, derivable=False)
def Wiener(x, y):
    """
    Wiener kernel.
    
    .. math::
        k(x, y) = \\min(x, y), \\quad x, y > 0
    
    A kernel representing a non-differentiable random walk starting at 0.
    
    Reference: Rasmussen and Williams (2006, p. 94).
    """
    with _patch_jax.skipifabstract():
        assert jnp.all(x >= 0)
        assert jnp.all(y >= 0)
    return jnp.minimum(x, y)

def _fracbrownian_derivable(H=1/2, K=1):
    return H == 1 and K == 1
    # TODO fails under tracing, return None if not concrete, maybe silence
    # derivability warnings under tracing

@kernel(forcekron=True, derivable=_fracbrownian_derivable)
def FracBrownian(x, y, H=1/2, K=1):
    """
    Bifractional Brownian motion kernel.
    
    .. math::
        k(x, y) = \\frac 1{2^K} \\big(
            (|x|^{2H} + |y|^{2H})^K - |x-y|^{2HK}
        \\big), \\quad H, K \\in (0, 1]
    
    For `H` = 1/2 (default) it is the Wiener kernel. For `H` in (0, 1/2) the
    increments are anticorrelated (strong oscillation), for `H` in (1/2, 1]
    the increments are correlated (tends to keep a slope).
    
    Reference: Houdré and Villa (2003).
    """
        
    # TODO I think the correlation between successive same step increments
    # is 2^(2H-1) - 1 in (-1/2, 1). Maybe add this to the docstring.
    
    with _patch_jax.skipifabstract():
        assert 0 < H <= 1, H
        assert 0 < K <= 1, K
    H2 = 2 * H
    return 1 / 2 ** K * ((jnp.abs(x) ** H2 + jnp.abs(y) ** H2) ** K - jnp.abs(x - y) ** (H2 * K))

# redefine derivatives of min and max because jax default is to yield 1/2
# when x == y, while I need 1 but consistently between min/max

@jax.custom_jvp
def _minimum(x, y):
    return jnp.minimum(x, y)

@_minimum.defjvp
def _minimum_jvp(primals, tangents):
    x, y = primals
    xdot, ydot = tangents
    return _minimum(x, y), jnp.where(x < y, xdot, ydot)

@jax.custom_jvp
def _maximum(x, y):
    return jnp.maximum(x, y)

@_maximum.defjvp
def _maximum_jvp(primals, tangents):
    x, y = primals
    xdot, ydot = tangents
    return _maximum(x, y), jnp.where(x >= y, xdot, ydot)

@kernel(derivable=1, forcekron=True)
def WienerIntegral(x, y):
    """
    Kernel for a process whose derivative is a Wiener process.
    
    .. math::
        k(x, y) = \\frac 12 a^2 \\left(b - \\frac a3 \\right),
        \\quad a = \\min(x, y), b = \\max(x, y)
    
    """
    
    # TODO can I generate this algorithmically for arbitrary integration order?
    # If I don't find a closed formula I can use sympy.
    
    with _patch_jax.skipifabstract():
        assert jnp.all(x >= 0)
        assert jnp.all(y >= 0)
    a = _minimum(x, y)
    b = _maximum(x, y)
    return 1/2 * a ** 2 * (b - a / 3)

@kernel(forcekron=True, derivable=False)
def OrnsteinUhlenbeck(x, y):
    """
    Ornstein-Uhlenbeck process kernel.
    
    .. math::
        k(x, y) = \\exp(-|x - y|) - \\exp(-(x + y)),
        \\quad x, y \\ge 0
    
    It is a random walk plus a negative feedback term that keeps the
    asymptotical variance constant. It is asymptotically stationary; often the
    name "Ornstein-Uhlenbeck" is given to the stationary part only, which here
    is provided as :class:`Expon`.
    
    """
    
    # TODO reference? look on wikipedia
    
    with _patch_jax.skipifabstract():
        assert jnp.all(x >= 0)
        assert jnp.all(y >= 0)
    return jnp.exp(-jnp.abs(x - y)) - jnp.exp(-(x + y))

@kernel(forcekron=True, derivable=False)
def BrownianBridge(x, y):
    """
    Brownian bridge kernel.
    
    .. math::
        k(x, y) = \\min(x, y) - xy,
        \\quad x, y \\in [0, 1]
    
    It is a Wiener process conditioned on being zero at x = 1.
    """
    
    # TODO reference? look on wikipedia
    
    # TODO can this have a Hurst index? I think the kernel would be
    # (t^2H(1-s) + s^2H(1-t) + s(1-t)^2H + t(1-s)^2H - (t+s) - |t-s|^2H + 2ts)/2
    # but I have to check if it is correct. (In new kernel FracBrownianBridge.)
    
    with _patch_jax.skipifabstract():
        assert jnp.all(x >= 0) and jnp.all(x <= 1)
        assert jnp.all(y >= 0) and jnp.all(y <= 1)
    return jnp.minimum(x, y) - x * y

def _stationaryfracbrownian_derivable(H=1/2):
    return H == 1

@stationarykernel(forcekron=True, derivable=_stationaryfracbrownian_derivable, input='signed')
def StationaryFracBrownian(delta, H=1/2):
    """
    Stationary fractional Brownian motion kernel.
    
    .. math::
        k(\\Delta) = \\frac 12 (|\\Delta+1|^{2H} + |\\Delta-1|^{2H} - 2|\\Delta|^{2H}),
        \\quad H \\in (0, 1]
        
    Reference: Gneiting and Schlather (2006, p. 272).
    """
    
    # TODO older reference, see [29] is GS06.
    
    with _patch_jax.skipifabstract():
        assert 0 < H <= 1, H
    H2 = 2 * H
    return 1/2 * (jnp.abs(delta + 1) ** H2 + jnp.abs(delta - 1) ** H2 - 2 * jnp.abs(delta) ** H2)
    
    # TODO is the bifractional version of this valid?




import functools
import jax

def value_and_ops(f, *ops, has_aux=False, **kw):
    if not ops:
        return f
    def fop(*args, **kw):
        y = f(*args, **kw)
        if has_aux:
            y, aux = y
            return y, (aux,)
        else:
            return y, ()
    def nextfop(fop):
        def nextfop(*args, **kw):
            y, aux = fop(*args, **kw)
            return y, aux + (y,)
        return nextfop
    for op in ops:
        fop = op(nextfop(fop), has_aux=True, **kw)
    @functools.wraps(f)
    def lastfop(*args, **kw):
        y, aux = fop(*args, **kw)
        if has_aux:
            return aux[1:] + (y,), aux[0]
        else:
            return aux + (y,)
    return lastfop

def f(x):
    print('ciao')
    return 1/2 * x ** 2, 'ciao'

fg = value_and_ops(f, jax.grad, jax.grad, jax.grad, has_aux=True)

print(fg(3.))



import numpy as np
from matplotlib import pyplot as plt
from numpy.polynomial import polynomial

import lsqfitgp as lgp

root = 1.5
mu = 2

root = float(root)
roots = np.repeat(root, mu)
pol = polynomial.polyfromroots(roots)
phi = -pol[1:] / pol[0]
acf = lgp._kernels._yule_walker_inv(phi)

fig, ax = plt.subplots(num='ar', clear=True)

x = np.arange(30)
y1 = root ** -x
ext = lgp._kernels._ar_evolve(phi, acf[1:], np.zeros(len(x) - len(acf)))
y2 = np.concatenate([acf, ext])

ax.plot(x, y1, label='power')
ax.plot(x, y2, label='YW')

ax.legend()
ax.set_yscale('log')

fig.show()



# lsqfitgp/docs/runcode.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""Run the python code in the rst files specified on the command line, but
only if the leading indentation is at least 4 spaces and there is a blank line
after the code block"""

import re
import sys
import textwrap
import warnings

import numpy as np
from matplotlib import pyplot as plt
import gvar

from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import TerminalFormatter

def pyprint(text):
    print(highlight(text, PythonLexer(), TerminalFormatter()))

sys.path.insert(0, '..')

pattern = re.compile(r'::\n\s*?\n(( {4,}.*\n)+)\s*?\n')

def runcode(file):
    with open(file, 'r') as stream:
        text = stream.read()
    
    # reset working environment
    plt.close('all')
    gvar.switch_gvar()
    np.random.seed(0)
    globals_dict = {}
    
    for match in pattern.finditer(text):
        codeblock = match.group(1)
        print(58 * '-' + '\n')
        code = textwrap.dedent(codeblock).strip()
        printcode = '\n'.join(f' {i + 1:2d}  ' + l for i, l in enumerate(code.split('\n')))
        pyprint(printcode)
        exec(code, globals_dict)

for file in sys.argv[1:]:
    s = f'*  running {file}  *'
    line = '*' * len(s)
    print('\n' + line + '\n' + s + '\n' + line)
    runcode(file)



from scipy import special
import numpy as np
from matplotlib import pyplot as plt

fig, axs = plt.subplots(2, 1, num='bessel', clear=True, figsize=[6.4, 7])

n = np.arange(100)
zeros = np.vectorize(special.jn_zeros, signature='(),()->(n)')(n, 4)

ax = axs[0]

for i, y in enumerate(zeros.T):
    ax.plot(n, y, label=f'{i+1}-th zero')

ax.legend()
ax.set_xlabel('$\\nu$')
ax.set_ylabel('Zero of $J_\\nu$')
ax.grid()

@np.vectorize
def scale(nu):
    lnu = np.floor(nu)
    rnu = np.ceil(nu)
    zl, = special.jn_zeros(lnu, 1)
    if lnu == rnu:
        return zl
    else:
        zr, = special.jn_zeros(rnu, 1)
        return zl + (nu - lnu) * (zr - zl) / (rnu - lnu)

def kernel(x, nu):
    x = np.abs(x) * (2 + nu/2)# scale(nu)
    return special.gamma(nu + 1) * 2 ** nu * x ** -nu * special.jv(nu, x)

nu = np.linspace(0, 10, 10)
x = np.linspace(0, 5, 100)
k = kernel(x, nu[:, None])

ax = axs[1]

for y, n in zip(k, nu):
    ax.plot(x, y, label=f'$\\nu$={n:.2g}')

ax.legend()
ax.set_xlabel('$x$')
ax.set_ylabel('$J_\\nu(x)$')
ax.grid()

fig.show()

d = np.arange(10)
nu = np.arange(10)
yi = np.vectorize(special.jvp)(nu[:, None], 0, 2 * d)
print(yi)

nu = np.stack([nu - 1e-3, nu + 1e-3], 1).reshape(-1)
y = np.vectorize(special.jvp)(nu[:, None], 0, 2 * d)
print(y)



import itertools

import numpy as np
from matplotlib import pyplot as plt
import gvar

import lsqfitgp as lgp

fig, ax = plt.subplots(num='zeta5', clear=True)

clist = [2] # coherence length > 0
vlist = [2] # smoothness parameter >= 0
alist = [1] # seasonal variance in [0, 1]
slist = [5] # correlation length > 0

x = np.linspace(0, 10, 1001)
for c, v, a, s in itertools.product(clist, vlist, alist, slist):
    covfun = a * lgp.Zeta(nu=v) * lgp.Matern(nu=v, scale=c) + (1 - a) * lgp.Matern(nu=v, scale=s)
    gp = lgp.GP(covfun)
    gp.addx(x, 0)
    p = gp.prior(0)
    y = gvar.sample(p, eps=1e-12)
    ax.plot(x, y, label=f'c={c}, v={v}, a={a}, s={s}')

ax.legend()

fig.show()



import jax
import functools
import jax.test_util

@jax.tree_util.register_pytree_node_class
class Decomp:
    
    def tree_flatten(self):
        return (self.l, self.a), None
    
    @classmethod
    def tree_unflatten(cls, aux_data, children):
        self = super().__new__(cls)
        self.l, self.a = children
        return self

    # use __new__ instead of __init__ because __init__ does not return anything
    @functools.partial(jax.jit, static_argnums=0)
    def __new__(cls, a):
        self = super().__new__(cls)
        # stops a's gradient now since we are going to define custom gradients
        # w.r.t. a anyway
        self.l = jax.scipy.linalg.cholesky(jax.lax.stop_gradient(a), lower=True)
        self.a = a
        return self
    
    def solve(self, b):
        return self._solve(self, b)
    
    @jax.custom_jvp
    @staticmethod # staticmethod otherwise custom_jvp won't see self
    @jax.jit
    def _solve(self, b):
        lb = jax.scipy.linalg.solve_triangular(self.l, b, lower=True)
        llb = jax.scipy.linalg.solve_triangular(self.l.T, lb, lower=False)
        return llb
    
    @_solve.defjvp
    @jax.jit
    def _solve_jvp(primals, tangents):
        self, b = primals
        self_dot, b_dot = tangents
        a_dot = self_dot.a
        primal = self.solve(b)
        tangent_a = -self.solve(a_dot) @ primal # -a^-1 ∂a a^-1 b
        tangent_b = self.solve(b_dot)
        return primal, tangent_a + tangent_b

key = jax.random.PRNGKey(0)
m = jax.random.normal(key, (5, 5))
_, key = jax.random.split(key)
b = jax.random.normal(key, (5,))
a = m @ m.T
dec = Decomp(a)
x = dec.solve(b)
print(jax.numpy.linalg.norm(a @ x - b, 2)) # -> 3.095427e-06

# rev mode does not work, dunno why
jax.test_util.check_grads(lambda b: dec.solve(b), (b,), order=2, modes='fwd')
jax.test_util.check_grads(lambda m: Decomp(m @ m.T).solve(b), (m,), order=2, modes='fwd')



# lsqfitgp/_special/_zeta.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools
import math

import jax
from jax import lax
from jax import numpy as jnp
from jax.scipy import special as jspecial

from .. import _patch_jax
from . import _gamma

def _hurwitz_zeta_series(m, x, a1, onlyeven=False, onlyodd=False, skipterm=None):
    """
    hurwitz zeta(s = m + x, a = 1 - a1) with integer m
    meant to be used with |x| ≤ 1/2, but no actual restriction
    assuming -S <= s <= 0 and |a1| <= 1/2 with S ~ some decade
    https://dlmf.nist.gov/25.11.E10
    """
    
    # decide number of terms to sum
    t = _patch_jax.float_type(m, x, a1)
    nmax = _hze_nmax(t)
    n = jnp.arange(nmax + 1)
    
    # make arguments broadcastable with n
    x = x[..., None]
    m = m[..., None]
    a1 = a1[..., None]
    if skipterm is not None:
        skipterm = skipterm[..., None]
    
    # compute pochhammer symbol, factorial and power terms
    nm = n + m
    ns1 = nm - 1 + x # = n + s - 1
    ns1_limit = jnp.where(ns1 == 0, 1, ns1) # pochhammer zero cancels zeta pole
    ns1_limit = jnp.where(ns1 == 1, 0, ns1_limit)
    # TODO this == 1 worries me, maybe sometimes it's violated, use shift
    factor = jnp.cumprod((ns1_limit * a1 / n).at[..., 0].set(1), -1, t)
    
    # handle tweaks to the series
    if onlyeven:
        sl = slice(None, None, 2)
    elif onlyodd:
        sl = slice(1, None, 2)
    if onlyeven or onlyodd:
        n = n[sl]
        nm = nm[..., sl]
        ns1 = ns1[..., sl]
        factor = factor[..., sl]
    if skipterm is not None:
        factor = jnp.where(n == skipterm, 0, factor)
    
    # compute zeta term
    zet = zeta(x, nm) # = zeta(n + s)
    zet_limit = jnp.where(ns1 == 0, 1, zet) # pole cancelled by pochhammer
    
    # sum series
    kw = dict(precision=lax.Precision.HIGHEST)
    series = jnp.matmul(factor[..., None, :], zet_limit[..., :, None], **kw)
    return series.squeeze((-2, -1))

def _hze_nmax(t):
    minz = 0.0037 # = min(2 gamma(s) / (2 pi)^s) for s <= 0
    return int(math.ceil(-math.log2(jnp.finfo(t).eps * minz)))

# @jax.jit
def hurwitz_zeta(s, a):
    """
    For 0 <= a <= 1 and -S <= s <= 0 with S not too large
    """
    s = jnp.asarray(s)
    a = jnp.asarray(a)
    
    cond = a < 1/2  # do a + 1 to bring a closer to 1
    a1 = jnp.where(cond, -a, 1. - a)
    zero = jnp.array(0)
    zeta = _hurwitz_zeta_series(zero, s, a1)  # https://dlmf.nist.gov/25.11.E10
    zeta += jnp.where(cond, a ** -s, 0)  # https://dlmf.nist.gov/25.11.E3
    return zeta
    
    # https://specialfunctions.juliamath.org/stable/functions_list/#SpecialFunctions.zeta

@functools.partial(jax.custom_jvp, nondiff_argnums=(1, 2))
# @functools.partial(jax.jit, static_argnums=(2,))
def periodic_zeta(x, s, imag=False):
    """
    compute F(x,s) = Li_s(e^2πix) for real s > 1, real x
    """
    
    x = jnp.asarray(x)
    s = jnp.asarray(s)
    
    # decide boundary for large/small s implementation
    t = _patch_jax.float_type(x, s)
    eps = jnp.finfo(t).eps
    nmax = 50
    larges = math.ceil(-math.log(eps) / math.log(nmax)) # 1/nmax^s < eps
    
    z_smalls = _periodic_zeta_smalls(x, s, imag)
    z_larges = _periodic_zeta_larges(x, s, nmax, imag)

    return jnp.where(s < larges, z_smalls, z_larges)
    
    # TODO rewrite without vectorization and then use vectorize, use switch
    # instead of where to avoid evaluating all branches all times

@periodic_zeta.defjvp
def _periodic_zeta_jvp(s, imag, p, t):
    x, = p
    xt, = t
    primal = periodic_zeta(x, s, imag)
    sgn = 1 if imag else -1
    tangent = 2 * jnp.pi * sgn * periodic_zeta(x, s - 1, not imag) * xt
    return primal, tangent

def _standard_x(x):
    """ bring x in [0, 1/2] by modulus and reflection """
    x %= 1
    neg = x > 1/2
    return neg, jnp.where(neg, 1 - x, x)

def _periodic_zeta_larges(x, s, nmax, imag):
    """ https://dlmf.nist.gov/25.13.E1 """

    t = _patch_jax.float_type(x, s)
    s = s.astype(t) # avoid n^s overflow with integer s
    n = jnp.arange(1, nmax + 1)
    neg, nx = _standard_x(n * x[..., None])
    func = jnp.sin if imag else jnp.cos
    terms = func(2 * jnp.pi * nx) / n ** s[..., None]
    if imag:
        terms *= jnp.where(neg, -1, 1)
    return jnp.sum(terms, -1)

def _periodic_zeta_smalls(x, s, imag):
    """
    https://dlmf.nist.gov/25.11.E10 and https://dlmf.nist.gov/25.11.E3 expanded
    into https://dlmf.nist.gov/25.13.E2
    """
    neg, x = _standard_x(x) # x in [0, 1/2]
    
    eps = jnp.finfo(_patch_jax.float_type(x, s)).eps
    s = jnp.where(s % 1, s, s * (1 + eps)) # avoid integer s

    s1 = 1 - s  # < 0
    q = -jnp.around(s1).astype(int)
    a = s1 + q
    # now s1 == -q + a with q integer >= 0 and |a| <= 1/2
    
    pi = (2 * jnp.pi) ** -s1
    gam = _gamma.gamma(s1)
    func = _sin_pi2 if imag else _cos_pi2
    pha = func(-q, a) # = sin or cos(π/2 s1), numerically accurate for small a
    hzs = 2 * _hurwitz_zeta_series(-q, a, -x, onlyeven=not imag, onlyodd=imag, skipterm=q)
    # hzs = ζ(s1,1+x) -/+ ζ(s1,1-x) but without the x^q term in the series
    pdiff = zeta_series_power_diff(x, q, a)
    # pdiff accurately handles the sum of the external power x^-s1 due to
    # 25.11.E3 with the q-th term (cancellation) with even q for the real part
    # and odd q for the imaginary part
    cancelcond = jnp.logical_and(imag, q % 2 == 1)
    cancelcond |= jnp.logical_and(not imag, q % 2 == 0)
    power = jnp.where(cancelcond, pdiff, x ** -s1)
    hz = power + hzs # = ζ(s1,x) -/+ ζ(s1,1-x)
    
    out = (pi * gam * pha) * hz
    if imag:
        out *= jnp.where(neg, -1, 1)
    return out

def _cos_pi2(n, x):
    """ compute cos(π/2 (n + x)) for n integer, accurate for small x """
    arg = -jnp.pi / 2 * x
    cos = jnp.where(n % 2, jnp.sin(arg), jnp.cos(arg))
    return cos * jnp.where(n // 2 % 2, -1, 1)

def _sin_pi2(n, x):
    return _cos_pi2(n - 1, x)

def zeta_series_power_diff(x, q, a):
    """
    compute x^q-a + (-1)^q * [q-th term of 2 * hurwitz_zeta_series(-q, a, x)]
    """
    pint = x ** q
    pz = jnp.where(q, 0, jnp.where(a, -1, 0)) # * 0^q = 0^q-a - 0^q
    pdif = jnp.where(x, jnp.expm1(-a * jnp.log(x)), pz) # * x^q = x^q-a - x^q
    gamincr = jnp.where(q, _gamma.gamma_incr(1 + q, -a), 0)
    # gamincr = Γ(1+q-a) / Γ(1+q)Γ(1-a)  -  1
    zz = zeta_zero(a) # = ζ(a) - ζ(0)
    qdif = 2 * (1 + gamincr) * zz - gamincr # = (q-th term) - (q-th term)|_a=0
    return pint * (pdif + qdif)

def zeta_zero(s):
    """
    Compute zeta(s) - zeta(0) for |s| < 1 accurately
    """
    
    # f(s) = zeta(s) - 1 / (s - 1)
    # I have the Taylor series of f(s)
    # zeta(s) - zeta(0) = f(s) + 1 / (s - 1) + 1/2 =
    # = f(s) + 1/(s-1) + 1 - 1 + 1/2 =
    # = f(s) - 1/2 + s/(s-1)
    
    t = _patch_jax.float_type(s)
    coef = jnp.array(_zeta_zero_coef, t).at[0].set(0)
    fact = jnp.cumprod(jnp.arange(coef.size).at[0].set(1), dtype=t)
    coef /= fact
    f = jnp.polyval(coef[::-1], s)
    return f + s / (s - 1)

_zeta_zero_coef = [         # = _gen_zeta_zero_coef(17)
    0.5,
    0.08106146679532726,
    -0.006356455908584851,
    -0.004711166862254448,
    0.002896811986292041,
    -0.00023290755845472455,
    -0.0009368251300509295,
    0.0008498237650016692,
    -0.00023243173551155957,
    -0.00033058966361229646,
    0.0005432341157797085,
    -0.00037549317290726367,
    -1.960353628101392e-05,
    0.00040724123256303315,
    -0.0005704920132817777,
    0.0003939270789812044,
    8.345880582550168e-05,
]

def _gen_zeta_zero_coef(n): # pragma: no cover
    """
    Compute first n derivatives of zeta(s) - 1/(s-1) at s = 0
    """
    import mpmath as mp
    with mp.workdps(32):
        func = lambda s: mp.zeta(s) - 1 / (s - 1)
        return [float(mp.diff(func, 0, k)) for k in range(n)]

# @jax.jit
def zeta(s, n=0):
    """ compute ζ(n + s) with integer n, accurate for even n < 0 and small s """
    return jnp.where(n + s >= 1, _zeta_right(s, n), _zeta_left(s, n))

def _zeta_right(s, n):
    return jspecial.zeta(n + s, 1)
    
    # jax's zeta implements a generic hurwitz algorithm, see the GSL
    # https://www.gnu.org/software/gsl/doc/html/specfunc.html#zeta-functions
    # to implement only the zeta in a possibly better way

def _zeta_left(s, n):
    # reflection formula https://dlmf.nist.gov/25.4.E1
    m = 1 - n
    x = -s
    # m + x = 1 - (n + s) = 1 - n - s
    mx = m + x # > 0
    logpi = -mx * jnp.log(2 * jnp.pi)
    cos = _cos_pi2(m, x) # = cos(π/2 (m + x)) but accurate for small x
    loggam = jspecial.gammaln(mx)
    zet = _zeta_right(x, m)
    
    # cancel zeta pole at 1
    cos = jnp.where(mx == 1, -jnp.pi / 2, cos)
    zet = jnp.where(mx == 1,           1, zet)
        
    return 2 * jnp.exp(logpi + loggam) * cos * zet

# TODO in general use jnp.piecewise, which only computes necessary cases, and
# lax.while_loop, to avoid unnecessary iterations. Note that piecewise requires
# traceable functions. Also, piecewise needs the input to be a single array,
# use lax.switch manually for multiple arguments.



import jax
from jax import numpy as jnp
import numpy as np

import lsqfitgp as lgp

def f(a, x):
    print(a)
    return 2 * x

g = jax.vmap(jax.grad(f, 1), in_axes=(None, 0))

print(g(lgp.StructuredArray(np.zeros(1, 'f8,f8')), jnp.ones(1)))

def f(x, *, a):
    print(a)
    return 3 * x

g = jax.vmap(jax.grad(f))

print(g(jnp.ones(1), a=lgp.StructuredArray(np.zeros(1, 'f8,f8'))))

def f(x):
    assert x['a'].shape == ()
    print(x.get('c'))
    return x['a'] + 2 * x['b']

g = jax.vmap(f)

print(g(dict(a=jnp.ones(5), b=jnp.arange(5.), c=lgp.StructuredArray(np.zeros(5, 'f8,f8')))))

u = jax.vmap(jax.grad(f))
print(u(dict(a=jnp.ones(5), b=jnp.arange(5.), c=lgp.StructuredArray(np.zeros(5, 'f8,f8')))))

def f(x):
    return x['f0'] + 2.3 * x['f1']

def g(x, f1):
    x['f1'] = f1
    return f(x)

h = jax.jacfwd(g, 1)

i = jax.vmap(h)

def j(x):
    return i(x, x['f1'])

print(j(lgp.StructuredArray(np.ones(5, 'f8,f8'))))



# lsqfitgp/examples/q.py
#
# Copyright (c) 2020, 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE Q.

    Where we extend a pattern of waves in space, but do not dare
    look too far from our data.

"""

import lsqfitgp as lgp
from lsqfitgp import _linalg
from matplotlib import pyplot as plt
from mpl_toolkits import mplot3d
import numpy as np
import gvar

xdata1d = np.linspace(-4, 4, 10)
xpred1d = np.linspace(-10, 10, 50)

def makexy(x1d, y1d):
    xy = np.empty((len(x1d), len(y1d)), dtype=[
        ('x', float),
        ('y', float)
    ])
    x, y = np.meshgrid(x1d, y1d)
    xy['x'] = x
    xy['y'] = y
    return xy
    
xdata = makexy(xdata1d, xdata1d)
xpred = makexy(xpred1d, xpred1d)
y = np.cos(xdata['x']) * np.cos(xdata['y'])

gp = lgp.GP(lgp.ExpQuad(scale=3, dim='x') * lgp.ExpQuad(scale=3, dim='y'))
gp.addx(xdata.reshape(-1), 'pere')
gp.addx(xpred.reshape(-1), 'banane')

print('fit...')
m, cov = gp.predfromdata({'pere': y.reshape(-1)}, 'banane', raw=True)

print('samples...')
sample = m + gp.decompose(cov, solver='chol', epsrel=1e-4).correlate(np.random.randn(len(m)))
sample = sample.reshape(xpred.shape)

print('plot...')
fig, ax = plt.subplots(num='q', clear=True, subplot_kw=dict(projection='3d', computed_zorder=False))

ax.scatter(xdata['x'].reshape(-1), xdata['y'].reshape(-1), y.reshape(-1), color='black', zorder=10)
ax.plot_surface(xpred['x'], xpred['y'], sample, alpha=0.9, cmap='viridis')
ax.view_init(elev=70, azim=110)

fig.show()



import itertools

import numpy as np
from matplotlib import pyplot as plt
import gvar

import lsqfitgp as lgp

fig, ax = plt.subplots(num='zeta6', clear=True)

x = np.linspace(0, 1, 1001)
f = lgp.Zeta(nu=0.3, derivable=True).diff(0, 1)
y = f(0, x)
ax.plot(x, y)

ax.legend()

fig.show()



# lsqfitgp/examples/u.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""

                            EXAMPLE U.

    Where we infer the temporal scale of a process assuming
    another process is correlated with its derivative.

"""

import lsqfitgp as lgp
from matplotlib import pyplot as plt
import numpy as np
import gvar

data_deriv = 1

time = np.linspace(-5, 5, 10)
x = np.empty(len(time), dtype=[
    ('time', float),
    ('label', int)
])
x['time'] = time
x['label'] = 1

data_error = 0.05
data_mean = np.cos(time)
data_mean += data_error * np.random.randn(*data_mean.shape)
data = gvar.gvar(data_mean, np.full_like(data_mean, data_error))

label_scale = 5
corr = lgp.ExpQuad(scale=label_scale)(0, 1)
print(f'corr = {corr:.3g}')

def makegp(params):
    kernel_time = lgp.ExpQuad(scale=params['time_scale'], dim='time')
    kernel_label = lgp.ExpQuad(scale=label_scale, dim='label')
    gp = lgp.GP(kernel_time * kernel_label)
    gp.addx(x, 'data', deriv=(data_deriv, 'time'))
    gp.addx(np.array([(0, 0)], dtype=x.dtype), 'fixed_point')
    return gp

prior = {
    'log(time_scale)': gvar.log(gvar.gvar(3, 2))
}
datadict = {'data': data, 'fixed_point': [gvar.gvar(0, 1e2)]}
params = lgp.empbayes_fit(prior, makegp, datadict, raises=False, jit=True).p
print('time_scale:', params['time_scale'])
gp = makegp(gvar.mean(params))

time_pred = np.linspace(-10, 10, 100)
xpred = np.empty((2, len(time_pred)), dtype=x.dtype)
xpred['time'] = time_pred
xpred['label'][0] = 0
xpred['label'][1] = 1
gp.addx(xpred[0], 0)
gp.addx(xpred[1], 1, deriv=(1, 'time'))

pred = gp.predfromdata(datadict, [0, 1])

fig, ax = plt.subplots(num='u', clear=True)

colors = dict()
for deriv in pred:
    m = gvar.mean(pred[deriv])
    s = gvar.sdev(pred[deriv])
    polys = ax.fill_between(time_pred, m - s, m + s, alpha=0.5, label=f'deriv {deriv}')
    colors[deriv] = polys.get_facecolor()[0]

for sample in gvar.raniter(pred, 3):
    for deriv in pred:
        ax.plot(time_pred, sample[deriv], color=colors[deriv])

ax.errorbar(time, gvar.mean(data), yerr=gvar.sdev(data), fmt='.', color=colors[data_deriv], alpha=1, label='data')

ax.legend(loc='best')
ax.set_xlabel('time')

fig.show()



import lsqfitgp as lgp
import numpy as np
import gvar
import jax

hyperprior = gvar.gvar(1, 1)

def makegp(p):
    gp = lgp.GP((0.01 + p ** 2) * lgp.White(), checkpos=False, checkfinite=False, solver='chol')
    gp.addx(0, 0)
    return gp

data = {0: 1}

fit = lgp.empbayes_fit(hyperprior, makegp, data, raises=False, verbosity=3, jit=True, method='gradient', mlkw=dict(direct_autodiff=True))
